# Copyright (c) 2015 Orange.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import copy

import mock

from oslo_config import cfg

from oslo_utils import uuidutils

from networking_bagpipe.agent.bagpipe_bgp_agent import BAGPIPE_NOTIFIER
from networking_bagpipe.agent.bagpipe_bgp_agent import BAGPIPE_NOTIFIERS

from networking_bagpipe.agent.bagpipe_bgp_agent import BaGPipeBGPAgent
from networking_bagpipe.agent.bagpipe_bgp_agent import BaGPipeBGPException

from networking_bagpipe.agent.bagpipe_bgp_agent import BGPVPN_L2
from networking_bagpipe.agent.bagpipe_bgp_agent import BGPVPN_L3
from networking_bagpipe.agent.bagpipe_bgp_agent import BGPVPN_NOTIFIER
from networking_bagpipe.agent.bagpipe_bgp_agent import BGPVPN_TYPES
from networking_bagpipe.agent.bagpipe_bgp_agent import BGPVPN_TYPES_MAP
from networking_bagpipe.agent.bagpipe_bgp_agent import EVPN
from networking_bagpipe.agent.bagpipe_bgp_agent import IPVPN
from networking_bagpipe.agent.bagpipe_bgp_agent import VPN_TYPES

from neutron.plugins.ml2.drivers.linuxbridge.agent.linuxbridge_neutron_agent \
    import LinuxBridgeManager
from neutron.plugins.ml2.drivers.openvswitch.agent import vlanmanager

from neutron.tests import base
from neutron.tests.unit.plugins.ml2.drivers.openvswitch.agent import (
    ovs_test_base)

from neutron_lib import constants as n_const

import logging

LOG = logging.getLogger(__name__)

PATCH_TUN_TO_MPLS_OFPORT = 1
PATCH_TUN_FROM_MPLS_OFPORT = 3
PATCH_TUN_OFPORTS = [PATCH_TUN_TO_MPLS_OFPORT, PATCH_TUN_FROM_MPLS_OFPORT]

PATCH_INT_TO_MPLS_OFPORT = 5
PATCH_INT_OFPORTS = [PATCH_INT_TO_MPLS_OFPORT]

PATCH_MPLS_FROM_TUN_OFPORT = 2
PATCH_MPLS_TO_TUN_OFPORT = 4
PATCH_MPLS_TO_INT_OFPORT = 6
PATCH_MPLS_OFPORTS = [PATCH_MPLS_FROM_TUN_OFPORT, PATCH_MPLS_TO_TUN_OFPORT,
                      PATCH_MPLS_TO_INT_OFPORT]

PORT10 = {'id': uuidutils.generate_uuid(),
          'mac_address': '00:00:de:ad:be:ef',
          'ip_address': '10.0.0.2'}

PORT11 = {'id': uuidutils.generate_uuid(),
          'mac_address': '00:00:de:ad:f0:0d',
          'ip_address': '10.0.0.3'}

NETWORK1 = {'id': uuidutils.generate_uuid(),
            'gateway_ip': '10.0.0.1'}

PORT20 = {'id': uuidutils.generate_uuid(),
          'mac_address': '00:00:de:ad:be:ef',
          'ip_address': '20.0.0.2'}

PORT21 = {'id': uuidutils.generate_uuid(),
          'mac_address': '00:00:de:ad:f0:0d',
          'ip_address': '20.0.0.3'}

NETWORK2 = {'id': uuidutils.generate_uuid(),
            'gateway_ip': '20.0.0.1'}

LOCAL_VLAN_MAP = {
    NETWORK1['id']: 31,
    NETWORK2['id']: 52
}

BAGPIPE_EVPN_RT1 = {'import_rt': ['BAGPIPE_EVPN:1'],
                    'export_rt': ['BAGPIPE_EVPN:1']}

BAGPIPE_EVPN_RT2 = {'import_rt': ['BAGPIPE_EVPN:2'],
                    'export_rt': ['BAGPIPE_EVPN:2']}

BGPVPN_EVPN_RT10 = {'import_rt': ['BGPVPN_EVPN:10'],
                    'export_rt': ['BGPVPN_EVPN:10']}

BGPVPN_EVPN_RT20 = {'import_rt': ['BGPVPN_EVPN:20'],
                    'export_rt': ['BGPVPN_EVPN:20']}

BGPVPN_IPVPN_RT100 = {'import_rt': ['BGPVPN_IPVPN:100'],
                      'export_rt': ['BGPVPN_IPVPN:100']}

BGPVPN_IPVPN_RT200 = {'import_rt': ['BGPVPN_IPVPN:200'],
                      'export_rt': ['BGPVPN_IPVPN:200']}


class DummyPort(object):
    def __init__(self, network, port, evpn=None, ipvpn=None):
        self.id = port['id']
        self.network_id = network['id']
        self.mac_address = port['mac_address']
        self.ip_address = port['ip_address']
        self.gateway_ip = network['gateway_ip']

        if evpn:
            self.evpn = evpn

        if ipvpn:
            self.ipvpn = ipvpn


class DummyVif(object):
    def __init__(self, ofport, port_name):
        self.ofport = ofport
        self.port_name = port_name


class DummyBGPVPN(object):
    def __init__(self, network, evpn=None, ipvpn=None, gateway_mac=None):
        self.network_id = network['id']

        if evpn:
            self.evpn = evpn

        if ipvpn:
            self.ipvpn = ipvpn

        if gateway_mac:
            self.gateway_mac = gateway_mac


class TestBaGPipeBGPAgentMixin(object):

    DUMMY_VIF10 = None
    DUMMY_VIF11 = None
    DUMMY_VIF20 = None
    DUMMY_VIF21 = None

    def _get_expected_local_port(self, network_id, port_id, vif_name):
        raise NotImplementedError

    def _get_expected_route_target(self, vpn_type, port, others_rts,
                                   others_last):
        if others_rts:
            others_import = others_rts.get('import_rt')
            others_export = others_rts.get('export_rt')
        else:
            others_import = []
            others_export = []

        if vpn_type in port:
            import_rt = list(port[vpn_type]['import_rt'])
            export_rt = list(port[vpn_type]['export_rt'])
        else:
            import_rt = []
            export_rt = []

        if others_last:
            import_rt += others_import
            export_rt += others_export
        else:
            import_rt = others_import + import_rt
            export_rt = others_export + export_rt

        return import_rt, export_rt

    def _mock_send_expected_call(self, vpn_type, port, vif, evpn2ipvpn=False,
                                 others_rts=None, others_last=False,
                                 fallback=None):
        network_id = port['network_id']

        vif_name = vif.port_name if vif else None
        local_port, linuxbr = self._get_expected_local_port(network_id,
                                                            port['id'],
                                                            vif_name)
        # Change local port if plugging evpn into ipvpn
        if evpn2ipvpn:
            local_port = dict(evpn=dict(id=network_id + '_evpn'))

        import_rt, export_rt = self._get_expected_route_target(vpn_type,
                                                               port,
                                                               others_rts,
                                                               others_last)

        expected_call = dict(vpn_instance_id=network_id + '_' + vpn_type,
                             vpn_type=vpn_type,
                             local_port=local_port,
                             mac_address=port['mac_address'],
                             ip_address=port['ip_address'],
                             gateway_ip=port['gateway_ip'],
                             import_rt=import_rt,
                             export_rt=export_rt)

        if linuxbr:
            expected_call.update(dict(
                linuxbr=LinuxBridgeManager.get_bridge_name(network_id))
            )
        if fallback:
            expected_call.update({'fallback': fallback})

        return mock.call(expected_call)

    def _check_network_attachments(self, network_id, expected_size,
                                   notifiers=BAGPIPE_NOTIFIERS,
                                   vpn_types=VPN_TYPES):
        if expected_size == 0:
            self.assertNotIn(network_id, self.agent.reg_attachments,
                             "Network %s expected to have no attachments left"
                             % network_id)
        else:
            self.assertIn(network_id, self.agent.reg_attachments)
            attachments = self.agent.reg_attachments[network_id]
            self.assertEqual(len(attachments), expected_size,
                             "Network attachments size not as expected")
            for attachment in attachments:
                for notifier in notifiers:
                    self.assertIn(notifier, attachment,
                                  "No %s details found in attachment %s" %
                                  (notifier, attachment))
                    for vpn_type in vpn_types:
                        self.assertIn(vpn_type, attachment[notifier],
                                      "No %s %s details found in "
                                      "attachment %s" % (notifier,
                                                         vpn_type,
                                                         attachment))

    # ----------------------------
    # BaGPipe RPC notifier tests |
    # ----------------------------
    def test_bagpipe_attach_single_port(self):
        dummy_port10 = DummyPort(NETWORK1, PORT10,
                                 evpn=BAGPIPE_EVPN_RT1).__dict__

        with mock.patch.object(self.agent,
                               'send_attach_local_port') as send_attach_fn:
            expected_calls = [
                self._mock_send_expected_call(EVPN,
                                              dummy_port10,
                                              self.DUMMY_VIF10)
            ]

            self.agent.bagpipe_port_attach(None, dummy_port10)

            send_attach_fn.assert_has_calls(expected_calls)

            self._check_network_attachments(NETWORK1['id'],
                                            1,
                                            [BAGPIPE_NOTIFIER],
                                            [EVPN])

    def test_bagpipe_attach_same_port_different_route_target(self):
        dummy_port10 = DummyPort(NETWORK1, PORT10,
                                 evpn=BAGPIPE_EVPN_RT1).__dict__
        dummy_port10bis = DummyPort(NETWORK1, PORT10,
                                    evpn=BAGPIPE_EVPN_RT2).__dict__

        with mock.patch.object(self.agent,
                               'send_attach_local_port') as send_attach_fn:
            expected_calls = [
                self._mock_send_expected_call(EVPN,
                                              dummy_port10,
                                              self.DUMMY_VIF10),
                self._mock_send_expected_call(EVPN,
                                              dummy_port10bis,
                                              self.DUMMY_VIF10)
            ]

            self.agent.bagpipe_port_attach(None, dummy_port10)
            self.agent.bagpipe_port_attach(None, dummy_port10bis)

            send_attach_fn.assert_has_calls(expected_calls)

            self._check_network_attachments(NETWORK1['id'],
                                            1,
                                            [BAGPIPE_NOTIFIER],
                                            [EVPN])

    def test_bagpipe_attach_multiple_ports_same_network(self):
        dummy_port10 = DummyPort(NETWORK1, PORT10,
                                 evpn=BAGPIPE_EVPN_RT1).__dict__
        dummy_port11 = DummyPort(NETWORK1, PORT11,
                                 evpn=BAGPIPE_EVPN_RT1).__dict__

        with mock.patch.object(self.agent,
                               'send_attach_local_port') as send_attach_fn:
            expected_calls = [
                self._mock_send_expected_call(EVPN,
                                              dummy_port10,
                                              self.DUMMY_VIF10),
                self._mock_send_expected_call(EVPN,
                                              dummy_port11,
                                              self.DUMMY_VIF11)
            ]

            self.agent.bagpipe_port_attach(None, dummy_port10)
            self.agent.bagpipe_port_attach(None, dummy_port11)

            send_attach_fn.assert_has_calls(expected_calls)

            self._check_network_attachments(NETWORK1['id'],
                                            2,
                                            [BAGPIPE_NOTIFIER],
                                            [EVPN])

    def test_bagpipe_attach_multiple_ports_different_networks(self):
        dummy_port10 = DummyPort(NETWORK1, PORT10,
                                 evpn=BAGPIPE_EVPN_RT1).__dict__
        dummy_port20 = DummyPort(NETWORK2, PORT20,
                                 evpn=BAGPIPE_EVPN_RT2).__dict__

        with mock.patch.object(self.agent,
                               'send_attach_local_port') as send_attach_fn:
            expected_calls = [
                self._mock_send_expected_call(EVPN,
                                              dummy_port10,
                                              self.DUMMY_VIF10),
                self._mock_send_expected_call(EVPN,
                                              dummy_port20,
                                              self.DUMMY_VIF20)
            ]

            self.agent.bagpipe_port_attach(None, dummy_port10)
            self.agent.bagpipe_port_attach(None, dummy_port20)

            send_attach_fn.assert_has_calls(expected_calls)

            for network_id in [NETWORK1['id'], NETWORK2['id']]:
                self._check_network_attachments(network_id,
                                                1,
                                                [BAGPIPE_NOTIFIER],
                                                [EVPN])

    def test_bagpipe_detach_single_port(self):
        dummy_port10 = DummyPort(NETWORK1, PORT10,
                                 evpn=BAGPIPE_EVPN_RT1).__dict__
        dummy_detach10 = dict(id=PORT10['id'], network_id=NETWORK1['id'])

        with mock.patch.object(self.agent,
                               'send_detach_local_port') as send_detach_fn:
            expected_calls = [
                self._mock_send_expected_call(EVPN,
                                              dummy_port10,
                                              self.DUMMY_VIF10)
            ]

            self.agent.bagpipe_port_attach(None, dummy_port10)
            self.agent.bagpipe_port_detach(None, dummy_detach10)

            send_detach_fn.assert_has_calls(expected_calls)

            self._check_network_attachments(NETWORK1['id'], 0)
            self.assertEqual(0, len(self.agent.reg_attachments),
                             "Registered attachments list must be empty: %s" %
                             self.agent.reg_attachments)

    def test_bagpipe_detach_multiple_ports_same_network(self):
        dummy_port10 = DummyPort(NETWORK1, PORT10,
                                 evpn=BAGPIPE_EVPN_RT1).__dict__
        dummy_detach10 = dict(id=PORT10['id'], network_id=NETWORK1['id'])

        dummy_port11 = DummyPort(NETWORK1, PORT11,
                                 evpn=BAGPIPE_EVPN_RT1).__dict__
        dummy_detach11 = dict(id=PORT11['id'], network_id=NETWORK1['id'])

        with mock.patch.object(self.agent,
                               'send_detach_local_port') as send_detach_fn:
            expected_calls = [
                self._mock_send_expected_call(EVPN,
                                              dummy_port10,
                                              self.DUMMY_VIF10),
                self._mock_send_expected_call(EVPN,
                                              dummy_port11,
                                              self.DUMMY_VIF11)
            ]

            # Attach 2 ports on network 1
            self.agent.bagpipe_port_attach(None, dummy_port10)
            self.agent.bagpipe_port_attach(None, dummy_port11)

            # Detach 1 port from network 1
            self.agent.bagpipe_port_detach(None, dummy_detach10)

            # Verify attachments list consistency
            self._check_network_attachments(NETWORK1['id'],
                                            1,
                                            [BAGPIPE_NOTIFIER],
                                            [EVPN])

            # Detach remaining port from network 1
            self.agent.bagpipe_port_detach(None, dummy_detach11)

            # Check if calls on BaGPipe BGP API are as expected
            send_detach_fn.assert_has_calls(expected_calls)

            # Verify attachments list consistency
            self._check_network_attachments(NETWORK1['id'], 0)
            self.assertEqual(0, len(self.agent.reg_attachments),
                             "Registered attachments list must be empty: %s" %
                             self.agent.reg_attachments)

    def test_bagpipe_detach_multiple_ports_different_networks(self):
        dummy_port10 = DummyPort(NETWORK1, PORT10,
                                 evpn=BAGPIPE_EVPN_RT1).__dict__
        dummy_detach10 = dict(id=PORT10['id'], network_id=NETWORK1['id'])

        dummy_port11 = DummyPort(NETWORK1, PORT11,
                                 evpn=BAGPIPE_EVPN_RT1).__dict__
        dummy_detach11 = dict(id=PORT11['id'], network_id=NETWORK1['id'])

        dummy_port20 = DummyPort(NETWORK2, PORT20,
                                 evpn=BAGPIPE_EVPN_RT2).__dict__
        dummy_detach20 = dict(id=PORT20['id'], network_id=NETWORK2['id'])

        dummy_port21 = DummyPort(NETWORK2, PORT21,
                                 evpn=BAGPIPE_EVPN_RT2).__dict__
        dummy_detach21 = dict(id=PORT21['id'], network_id=NETWORK2['id'])

        with mock.patch.object(self.agent,
                               'send_detach_local_port') as send_detach_fn:
            expected_calls = [
                self._mock_send_expected_call(EVPN,
                                              dummy_port10,
                                              self.DUMMY_VIF10),
                self._mock_send_expected_call(EVPN,
                                              dummy_port20,
                                              self.DUMMY_VIF20),
                self._mock_send_expected_call(EVPN,
                                              dummy_port11,
                                              self.DUMMY_VIF11),
                self._mock_send_expected_call(EVPN,
                                              dummy_port21,
                                              self.DUMMY_VIF21)
            ]

            # Attach 2 ports on network 1
            self.agent.bagpipe_port_attach(None, dummy_port10)
            self.agent.bagpipe_port_attach(None, dummy_port11)

            # Attach 2 ports on network 2
            self.agent.bagpipe_port_attach(None, dummy_port20)
            self.agent.bagpipe_port_attach(None, dummy_port21)

            # Detach 1 port from each network
            self.agent.bagpipe_port_detach(None, dummy_detach10)
            self.agent.bagpipe_port_detach(None, dummy_detach20)

            # Verify attachments list consistency
            for network_id in [NETWORK1['id'], NETWORK2['id']]:
                self._check_network_attachments(network_id,
                                                1,
                                                [BAGPIPE_NOTIFIER],
                                                [EVPN])

            # Detach remaining port from each network
            self.agent.bagpipe_port_detach(None, dummy_detach11)
            self.agent.bagpipe_port_detach(None, dummy_detach21)

            # Check if calls on BaGPipe BGP API are as expected
            send_detach_fn.assert_has_calls(expected_calls)

            # Verify attachments list consistency
            for network_id in [NETWORK1['id'], NETWORK2['id']]:
                self._check_network_attachments(network_id, 0)

            self.assertEqual(0, len(self.agent.reg_attachments),
                             "Registered attachments list must be empty: %s" %
                             self.agent.reg_attachments)

    # ----------------------------
    # BGP VPN RPC notifier tests |
    # ----------------------------
    def test_update_bgpvpn_no_plugged_ports(self):
        dummy_bgpvpn1 = DummyBGPVPN(NETWORK1,
                                    evpn=BGPVPN_EVPN_RT10).__dict__

        with mock.patch.object(self.agent,
                               'send_attach_local_port') as send_attach_fn:
            self.agent.update_bgpvpn(None, dummy_bgpvpn1)

            self.assertEqual(0, send_attach_fn.call_count,
                             "Send attach mustn't be called")

    def test_update_bgpvpn_already_plugged_ports(self):
        dummy_port10 = DummyPort(NETWORK1, PORT10).__dict__
        dummy_port11 = DummyPort(NETWORK1, PORT11).__dict__

        dummy_bgpvpn1 = DummyBGPVPN(NETWORK1,
                                    ipvpn=BGPVPN_IPVPN_RT100).__dict__

        with mock.patch.object(self.agent,
                               'send_attach_local_port') as send_attach_fn:
            expected_calls = [
                self._mock_send_expected_call(IPVPN,
                                              dummy_port10,
                                              self.DUMMY_VIF10,
                                              others_rts=BGPVPN_IPVPN_RT100),
                self._mock_send_expected_call(IPVPN,
                                              dummy_port11,
                                              self.DUMMY_VIF11,
                                              others_rts=BGPVPN_IPVPN_RT100)
            ]

            self.agent.bgpvpn_port_attach(None, dummy_port10)
            self.agent.bgpvpn_port_attach(None, dummy_port11)

            # Verify attachments list consistency
            self._check_network_attachments(NETWORK1['id'], 2,
                                            [BGPVPN_NOTIFIER])

            self.agent.update_bgpvpn(None, dummy_bgpvpn1)

            send_attach_fn.assert_has_calls(expected_calls)

            # Verify attachments list consistency
            self._check_network_attachments(NETWORK1['id'], 2,
                                            [BGPVPN_NOTIFIER])

    def test_update_bgpvpn_same_vpn_types(self):
        dummy_port10 = DummyPort(NETWORK1, PORT10,
                                 evpn=BGPVPN_EVPN_RT10).__dict__

        dummy_port10bis = DummyPort(NETWORK1, PORT10,
                                    evpn=BGPVPN_EVPN_RT20).__dict__

        evpn_rts = ({k: BGPVPN_EVPN_RT10[k] + BGPVPN_EVPN_RT20[k]
                    for k in ['import_rt', 'export_rt']})

        dummy_bgpvpn1 = DummyBGPVPN(NETWORK1,
                                    evpn=evpn_rts).__dict__

        with mock.patch.object(self.agent,
                               'send_attach_local_port') as send_attach_fn:
            expected_calls = [
                self._mock_send_expected_call(EVPN,
                                              dummy_port10,
                                              self.DUMMY_VIF10),
                self._mock_send_expected_call(EVPN,
                                              dummy_port10bis,
                                              self.DUMMY_VIF10,
                                              others_rts=BGPVPN_EVPN_RT10)
            ]

            self.agent.bgpvpn_port_attach(None, dummy_port10)

            self.agent.update_bgpvpn(None, dummy_bgpvpn1)

            send_attach_fn.assert_has_calls(expected_calls)

            self._check_network_attachments(NETWORK1['id'],
                                            1,
                                            [BGPVPN_NOTIFIER])

    def test_update_bgpvpn_different_vpn_types(self):
        dummy_port10 = DummyPort(NETWORK1, PORT10,
                                 evpn=BGPVPN_EVPN_RT10).__dict__

        dummy_port10bis = DummyPort(NETWORK1, PORT10,
                                    evpn=BGPVPN_EVPN_RT10,
                                    ipvpn=BGPVPN_IPVPN_RT100).__dict__

        dummy_bgpvpn1 = DummyBGPVPN(NETWORK1,
                                    evpn=BGPVPN_EVPN_RT10,
                                    ipvpn=BGPVPN_IPVPN_RT100).__dict__

        with mock.patch.object(self.agent,
                               'send_attach_local_port') as send_attach_fn:
            expected_calls = [
                self._mock_send_expected_call(EVPN,
                                              dummy_port10,
                                              self.DUMMY_VIF10),
                self._mock_send_expected_call(EVPN,
                                              dummy_port10bis,
                                              self.DUMMY_VIF10),
                self._mock_send_expected_call(IPVPN,
                                              dummy_port10bis,
                                              self.DUMMY_VIF10,
                                              evpn2ipvpn=True)
            ]

            self.agent.bgpvpn_port_attach(None, dummy_port10)

            self.agent.update_bgpvpn(None, dummy_bgpvpn1)

            send_attach_fn.assert_has_calls(expected_calls)

            self._check_network_attachments(NETWORK1['id'],
                                            1,
                                            [BGPVPN_NOTIFIER])

    def test_delete_bgpvpn_remaining_plugged_ports(self):
        dummy_port10 = DummyPort(NETWORK1, PORT10,
                                 evpn=BGPVPN_EVPN_RT10).__dict__
        dummy_port11 = DummyPort(NETWORK1, PORT11,
                                 evpn=BGPVPN_EVPN_RT10).__dict__

        dummy_bgpvpn1 = DummyBGPVPN(NETWORK1,
                                    evpn=BGPVPN_EVPN_RT10).__dict__

        with mock.patch.object(self.agent,
                               'send_detach_local_port') as send_detach_fn:
            expected_calls = [
                self._mock_send_expected_call(EVPN,
                                              dummy_port10,
                                              self.DUMMY_VIF10),
                self._mock_send_expected_call(EVPN,
                                              dummy_port11,
                                              self.DUMMY_VIF11)
            ]

            self.agent.bgpvpn_port_attach(None, dummy_port10)
            self.agent.bgpvpn_port_attach(None, dummy_port11)

            # Verify attachments list consistency
            self._check_network_attachments(NETWORK1['id'],
                                            2,
                                            [BGPVPN_NOTIFIER])

            self.agent.delete_bgpvpn(None, dummy_bgpvpn1)

            send_detach_fn.assert_has_calls(expected_calls)

            # Verify attachments list consistency
            self._check_network_attachments(NETWORK1['id'],
                                            2,
                                            [BGPVPN_NOTIFIER])

    def test_delete_bgpvpn_no_plugged_ports(self):
        dummy_bgpvpn1 = DummyBGPVPN(NETWORK1,
                                    evpn=BGPVPN_EVPN_RT10).__dict__

        with mock.patch.object(self.agent,
                               'send_detach_local_port') as send_detach_fn:
            self.agent.delete_bgpvpn(None, dummy_bgpvpn1)

            self.assertEqual(0, send_detach_fn.call_count,
                             "Send detach mustn't be called")

    def test_delete_bgpvpn_had_plugged_ports(self):
        dummy_port10 = DummyPort(NETWORK1, PORT10,
                                 ipvpn=BGPVPN_EVPN_RT10).__dict__
        dummy_detach10 = dict(id=PORT10['id'], network_id=NETWORK1['id'])

        dummy_port11 = DummyPort(NETWORK1, PORT11,
                                 ipvpn=BGPVPN_EVPN_RT10).__dict__
        dummy_detach11 = dict(id=PORT11['id'], network_id=NETWORK1['id'])

        dummy_bgpvpn1 = DummyBGPVPN(NETWORK1,
                                    ipvpn=BGPVPN_EVPN_RT10).__dict__

        with mock.patch.object(self.agent,
                               'send_detach_local_port') as send_detach_fn:

            self.agent.bgpvpn_port_attach(None, dummy_port10)
            self.agent.bgpvpn_port_attach(None, dummy_port11)

            self.agent.bgpvpn_port_detach(None, dummy_detach10)
            self.agent.bgpvpn_port_detach(None, dummy_detach11)

            self.assertEqual(2, send_detach_fn.call_count,
                             "Send detach must be called twice")

            self.assertEqual(0, len(self.agent.reg_attachments),
                             "Registered attachments list must be empty: %s" %
                             self.agent.reg_attachments)

            send_detach_fn.reset_mock()
            self.agent.delete_bgpvpn(None, dummy_bgpvpn1)

            self.assertEqual(0, send_detach_fn.call_count,
                             "Send detach ustn't be called")

    def _test_bgpvpn_attach_single_port(self, bgpvpn_type, bgpvpn_rts):
        mapped_type = (BGPVPN_TYPES_MAP[bgpvpn_type] if bgpvpn_type in
                       BGPVPN_TYPES else bgpvpn_type)
        bgpvpn_info = {mapped_type: bgpvpn_rts}
        dummy_port10 = DummyPort(NETWORK1, PORT10, **bgpvpn_info).__dict__

        with mock.patch.object(self.agent,
                               'send_attach_local_port') as send_attach_fn:
            expected_calls = [
                self._mock_send_expected_call(mapped_type,
                                              dummy_port10,
                                              self.DUMMY_VIF10)
            ]

            self.agent.bgpvpn_port_attach(None, dummy_port10)

            send_attach_fn.assert_has_calls(expected_calls)

            self._check_network_attachments(NETWORK1['id'],
                                            1,
                                            [BGPVPN_NOTIFIER],
                                            [mapped_type])

    def test_bgpvpn_attach_single_port_l3_bgpvpn1(self):
        self._test_bgpvpn_attach_single_port(BGPVPN_L3, BGPVPN_IPVPN_RT100)

    def test_bgpvpn_attach_single_port_l3_bgpvpn2(self):
        self._test_bgpvpn_attach_single_port(IPVPN, BGPVPN_IPVPN_RT100)

    def test_bgpvpn_attach_single_port_bgpvpn1(self):
        self._test_bgpvpn_attach_single_port(BGPVPN_L2, BGPVPN_EVPN_RT10)

    def test_bgpvpn_attach_single_port_bgpvpn2(self):
        self._test_bgpvpn_attach_single_port(EVPN, BGPVPN_EVPN_RT10)

    def test_bgpvpn_attach_same_port_different_bgpvpn(self):
        dummy_port10 = DummyPort(NETWORK1, PORT10,
                                 evpn=BGPVPN_EVPN_RT10).__dict__
        dummy_port10bis = DummyPort(NETWORK1, PORT10,
                                    evpn=BGPVPN_EVPN_RT10,
                                    ipvpn=BGPVPN_IPVPN_RT100).__dict__

        with mock.patch.object(self.agent,
                               'send_attach_local_port') as send_attach_fn:
            expected_calls = [
                self._mock_send_expected_call(EVPN,
                                              dummy_port10,
                                              self.DUMMY_VIF10),
                self._mock_send_expected_call(EVPN,
                                              dummy_port10bis,
                                              self.DUMMY_VIF10),
                self._mock_send_expected_call(IPVPN,
                                              dummy_port10bis,
                                              self.DUMMY_VIF10,
                                              evpn2ipvpn=True)
            ]

            self.agent.bgpvpn_port_attach(None, dummy_port10)

            self._check_network_attachments(NETWORK1['id'],
                                            1,
                                            [BGPVPN_NOTIFIER],
                                            [EVPN])

            self.agent.bgpvpn_port_attach(None, dummy_port10bis)

            send_attach_fn.assert_has_calls(expected_calls)

            self._check_network_attachments(NETWORK1['id'],
                                            1,
                                            [BGPVPN_NOTIFIER])

    def test_bgpvpn_attach_single_port_multiple_bgpvpns(self):
        dummy_port10 = DummyPort(NETWORK1, PORT10,
                                 evpn=BGPVPN_EVPN_RT10,
                                 ipvpn=BGPVPN_IPVPN_RT100).__dict__

        with mock.patch.object(self.agent,
                               'send_attach_local_port') as send_attach_fn:
            expected_calls = [
                self._mock_send_expected_call(EVPN,
                                              dummy_port10,
                                              self.DUMMY_VIF10),
                self._mock_send_expected_call(IPVPN,
                                              dummy_port10,
                                              self.DUMMY_VIF10,
                                              evpn2ipvpn=True)
            ]

            self.agent.bgpvpn_port_attach(None, dummy_port10)

            send_attach_fn.assert_has_calls(expected_calls)

            self._check_network_attachments(NETWORK1['id'],
                                            1,
                                            [BGPVPN_NOTIFIER],
                                            [EVPN, IPVPN])

    def test_bgpvpn_attach_multiple_ports_same_bgpvpn(self):
        dummy_port10 = DummyPort(NETWORK1, PORT10,
                                 ipvpn=BGPVPN_IPVPN_RT100).__dict__
        dummy_port11 = DummyPort(NETWORK1, PORT11,
                                 ipvpn=BGPVPN_IPVPN_RT100).__dict__

        with mock.patch.object(self.agent,
                               'send_attach_local_port') as send_attach_fn:
            expected_calls = [
                self._mock_send_expected_call(IPVPN,
                                              dummy_port10,
                                              self.DUMMY_VIF10),
                self._mock_send_expected_call(IPVPN,
                                              dummy_port11,
                                              self.DUMMY_VIF11)
            ]

            self.agent.bgpvpn_port_attach(None, dummy_port10)
            self.agent.bgpvpn_port_attach(None, dummy_port11)

            send_attach_fn.assert_has_calls(expected_calls)

            self._check_network_attachments(NETWORK1['id'],
                                            2,
                                            [BGPVPN_NOTIFIER],
                                            [IPVPN])

    def test_bgpvpn_attach_multiple_ports_different_bgpvpns(self):
        dummy_port10 = DummyPort(NETWORK1, PORT10,
                                 ipvpn=BGPVPN_IPVPN_RT100).__dict__
        dummy_port20 = DummyPort(NETWORK2, PORT20,
                                 ipvpn=BGPVPN_IPVPN_RT200).__dict__

        with mock.patch.object(self.agent,
                               'send_attach_local_port') as send_attach_fn:
            expected_calls = [
                self._mock_send_expected_call(IPVPN,
                                              dummy_port10,
                                              self.DUMMY_VIF10),
                self._mock_send_expected_call(IPVPN,
                                              dummy_port20,
                                              self.DUMMY_VIF20)
            ]

            self.agent.bgpvpn_port_attach(None, dummy_port10)
            self.agent.bgpvpn_port_attach(None, dummy_port20)

            send_attach_fn.assert_has_calls(expected_calls)

            for network_id in [NETWORK1['id'], NETWORK2['id']]:
                self._check_network_attachments(network_id,
                                                1,
                                                [BGPVPN_NOTIFIER],
                                                [IPVPN])

    def _test_bgpvpn_detach_single_port(self, bgpvpn_type, bgpvpn_rts):
        bgpvpn_info = {bgpvpn_type: bgpvpn_rts}
        dummy_port10 = DummyPort(NETWORK1, PORT10, **bgpvpn_info).__dict__
        dummy_detach10 = dict(id=PORT10['id'], network_id=NETWORK1['id'])

        with mock.patch.object(self.agent,
                               'send_detach_local_port') as send_detach_fn:
            expected_calls = [
                self._mock_send_expected_call(bgpvpn_type,
                                              dummy_port10,
                                              self.DUMMY_VIF10)
            ]

            self.agent.bgpvpn_port_attach(None, dummy_port10)
            self.agent.bgpvpn_port_detach(None, dummy_detach10)

            send_detach_fn.assert_has_calls(expected_calls)

            # Verify attachments list consistency
            self._check_network_attachments(NETWORK1['id'], 0)

            self.assertEqual(0, len(self.agent.reg_attachments),
                             "Registered attachments list must be empty: %s" %
                             self.agent.reg_attachments)

    def test_bgpvpn_detach_single_port_l3_bgpvpn(self):
        self._test_bgpvpn_detach_single_port(IPVPN, BGPVPN_IPVPN_RT100)

    def test_bgpvpn_detach_single_port_bgpvpn(self):
        self._test_bgpvpn_detach_single_port(EVPN, BGPVPN_EVPN_RT10)

    def test_bgpvpn_detach_single_port_multiple_bgpvpns(self):
        dummy_port10 = DummyPort(NETWORK1, PORT10,
                                 evpn=BGPVPN_EVPN_RT10,
                                 ipvpn=BGPVPN_IPVPN_RT100).__dict__
        dummy_detach10 = dict(id=PORT10['id'], network_id=NETWORK1['id'])

        with mock.patch.object(self.agent,
                               'send_detach_local_port') as send_detach_fn:
            expected_calls = [
                self._mock_send_expected_call(IPVPN,
                                              dummy_port10,
                                              self.DUMMY_VIF10,
                                              evpn2ipvpn=True),
                self._mock_send_expected_call(EVPN,
                                              dummy_port10,
                                              self.DUMMY_VIF10)
            ]

            self.agent.bgpvpn_port_attach(None, dummy_port10)
            self.agent.bgpvpn_port_detach(None, dummy_detach10)

            send_detach_fn.assert_has_calls(expected_calls)

            # Verify attachments list consistency
            self._check_network_attachments(NETWORK1['id'], 0)

            self.assertEqual(0, len(self.agent.reg_attachments),
                             "Registered attachments list must be empty: %s" %
                             self.agent.reg_attachments)

    def test_bgpvpn_detach_multiple_ports_same_bgpvpn(self):
        dummy_port10 = DummyPort(NETWORK1, PORT10,
                                 ipvpn=BGPVPN_IPVPN_RT100).__dict__
        dummy_detach10 = dict(id=PORT10['id'], network_id=NETWORK1['id'])

        dummy_port11 = DummyPort(NETWORK1, PORT11,
                                 ipvpn=BGPVPN_IPVPN_RT100).__dict__
        dummy_detach11 = dict(id=PORT11['id'], network_id=NETWORK1['id'])

        with mock.patch.object(self.agent,
                               'send_detach_local_port') as send_detach_fn:
            expected_calls = [
                self._mock_send_expected_call(IPVPN,
                                              dummy_port10,
                                              self.DUMMY_VIF10),
                self._mock_send_expected_call(IPVPN,
                                              dummy_port11,
                                              self.DUMMY_VIF11)
            ]

            # Attach 2 ports on BGP VPN L3 1
            self.agent.bgpvpn_port_attach(None, dummy_port10)
            self.agent.bgpvpn_port_attach(None, dummy_port11)

            # Detach 1 port from BGP VPN L3 1
            self.agent.bgpvpn_port_detach(None, dummy_detach10)

            # Detach remaining port from BGP VPN L3 1
            self.agent.bgpvpn_port_detach(None, dummy_detach11)

            # Check if calls on BaGPipe BGP API are as expected
            send_detach_fn.assert_has_calls(expected_calls)

            # Verify attachments list consistency
            self._check_network_attachments(NETWORK1['id'], 0)

            self.assertEqual(0, len(self.agent.reg_attachments),
                             "Registered attachments list must be empty: %s" %
                             self.agent.reg_attachments)

    def test_bgpvpn_detach_multiple_ports_different_bgpvpns(self):
        dummy_port10 = DummyPort(NETWORK1, PORT10,
                                 ipvpn=BGPVPN_IPVPN_RT100).__dict__
        dummy_detach10 = dict(id=PORT10['id'], network_id=NETWORK1['id'])

        dummy_port20 = DummyPort(NETWORK2, PORT20,
                                 ipvpn=BGPVPN_IPVPN_RT200).__dict__
        dummy_detach20 = dict(id=PORT20['id'], network_id=NETWORK2['id'])

        with mock.patch.object(self.agent,
                               'send_detach_local_port') as send_detach_fn:
            expected_calls = [
                self._mock_send_expected_call(IPVPN,
                                              dummy_port10,
                                              self.DUMMY_VIF10),
                self._mock_send_expected_call(IPVPN,
                                              dummy_port20,
                                              self.DUMMY_VIF20)
            ]

            self.agent.bgpvpn_port_attach(None, dummy_port10)
            self.agent.bgpvpn_port_attach(None, dummy_port20)

            self.agent.bgpvpn_port_detach(None, dummy_detach10)
            self.agent.bgpvpn_port_detach(None, dummy_detach20)

            send_detach_fn.assert_has_calls(expected_calls)

            # Verify attachments list consistency
            for network_id in [NETWORK1['id'], NETWORK2['id']]:
                self._check_network_attachments(network_id, 0)

            self.assertEqual(0, len(self.agent.reg_attachments),
                             "Registered attachments list must be empty: %s" %
                             self.agent.reg_attachments)

    # -------------------------------------------
    # Multiple simultaneous RPC notifiers tests |
    # -------------------------------------------
    def test_multiple_attach_single_port_evpns(self):
        bagpipe_port10 = DummyPort(NETWORK1, PORT10,
                                   evpn=BAGPIPE_EVPN_RT1).__dict__
        bgpvpn_port10 = DummyPort(NETWORK1, PORT10,
                                  evpn=BGPVPN_EVPN_RT10).__dict__

        with mock.patch.object(self.agent,
                               'send_attach_local_port') as send_attach_fn:
            expected_calls = [
                self._mock_send_expected_call(EVPN,
                                              bagpipe_port10,
                                              self.DUMMY_VIF10),
                self._mock_send_expected_call(EVPN,
                                              bgpvpn_port10,
                                              self.DUMMY_VIF10,
                                              others_rts=BAGPIPE_EVPN_RT1)
            ]

            self.agent.bagpipe_port_attach(None, bagpipe_port10)
            self.agent.bgpvpn_port_attach(None, bgpvpn_port10)

            send_attach_fn.assert_has_calls(expected_calls)

            self._check_network_attachments(NETWORK1['id'],
                                            1,
                                            [BAGPIPE_NOTIFIER,
                                             BGPVPN_NOTIFIER],
                                            [EVPN])

    def test_multiple_attach_single_port_different_vpns(self):
        bagpipe_port10 = DummyPort(NETWORK1, PORT10,
                                   evpn=BAGPIPE_EVPN_RT1).__dict__
        bgpvpn_port10 = DummyPort(NETWORK1, PORT10,
                                  ipvpn=BGPVPN_IPVPN_RT100).__dict__

        with mock.patch.object(self.agent,
                               'send_attach_local_port') as send_attach_fn:
            expected_calls = [
                self._mock_send_expected_call(EVPN,
                                              bagpipe_port10,
                                              self.DUMMY_VIF10),
                self._mock_send_expected_call(EVPN,
                                              bagpipe_port10,
                                              self.DUMMY_VIF10),
                self._mock_send_expected_call(IPVPN,
                                              bgpvpn_port10,
                                              self.DUMMY_VIF10,
                                              evpn2ipvpn=True)
            ]

            self.agent.bagpipe_port_attach(None, bagpipe_port10)
            self.agent.bgpvpn_port_attach(None, bgpvpn_port10)

            send_attach_fn.assert_has_calls(expected_calls)

            self._check_network_attachments(NETWORK1['id'],
                                            1,
                                            [BAGPIPE_NOTIFIER],
                                            [EVPN])
            self._check_network_attachments(NETWORK1['id'],
                                            1,
                                            [BGPVPN_NOTIFIER],
                                            [IPVPN])

    def test_multiple_attach_multiple_ports_same_evpns(self):
        bagpipe_port10 = DummyPort(NETWORK1, PORT10,
                                   evpn=BAGPIPE_EVPN_RT1).__dict__
        bgpvpn_port10 = DummyPort(NETWORK1, PORT10,
                                  evpn=BGPVPN_EVPN_RT10).__dict__

        bagpipe_port11 = DummyPort(NETWORK1, PORT11,
                                   evpn=BAGPIPE_EVPN_RT1).__dict__
        bgpvpn_port11 = DummyPort(NETWORK1, PORT11,
                                  evpn=BGPVPN_IPVPN_RT100).__dict__

        with mock.patch.object(self.agent,
                               'send_attach_local_port') as send_attach_fn:
            expected_calls = [
                self._mock_send_expected_call(EVPN,
                                              bagpipe_port10,
                                              self.DUMMY_VIF10),
                self._mock_send_expected_call(EVPN,
                                              bgpvpn_port10,
                                              self.DUMMY_VIF10,
                                              others_rts=BAGPIPE_EVPN_RT1),
                self._mock_send_expected_call(EVPN,
                                              bagpipe_port11,
                                              self.DUMMY_VIF11),
                self._mock_send_expected_call(EVPN,
                                              bgpvpn_port11,
                                              self.DUMMY_VIF11,
                                              others_rts=BAGPIPE_EVPN_RT1)
            ]

            self.agent.bagpipe_port_attach(None, bagpipe_port10)
            self.agent.bgpvpn_port_attach(None, bgpvpn_port10)

            self.agent.bagpipe_port_attach(None, bagpipe_port11)
            self.agent.bgpvpn_port_attach(None, bgpvpn_port11)

            send_attach_fn.assert_has_calls(expected_calls)

            self._check_network_attachments(NETWORK1['id'],
                                            2,
                                            BAGPIPE_NOTIFIERS,
                                            [EVPN])

    def test_multiple_attach_multiple_ports_different_evpns(self):
        bagpipe_port10 = DummyPort(NETWORK1, PORT10,
                                   evpn=BAGPIPE_EVPN_RT1).__dict__
        bgpvpn_port10 = DummyPort(NETWORK1, PORT10,
                                  evpn=BGPVPN_EVPN_RT10).__dict__

        bagpipe_port20 = DummyPort(NETWORK2, PORT20,
                                   evpn=BAGPIPE_EVPN_RT2).__dict__
        bgpvpn_port20 = DummyPort(NETWORK2, PORT20,
                                  evpn=BGPVPN_EVPN_RT20).__dict__

        with mock.patch.object(self.agent,
                               'send_attach_local_port') as send_attach_fn:
            expected_calls = [
                self._mock_send_expected_call(EVPN,
                                              bagpipe_port10,
                                              self.DUMMY_VIF10),
                self._mock_send_expected_call(EVPN,
                                              bgpvpn_port10,
                                              self.DUMMY_VIF10,
                                              others_rts=BAGPIPE_EVPN_RT1),
                self._mock_send_expected_call(EVPN,
                                              bagpipe_port20,
                                              self.DUMMY_VIF20),
                self._mock_send_expected_call(EVPN,
                                              bgpvpn_port20,
                                              self.DUMMY_VIF20,
                                              others_rts=BAGPIPE_EVPN_RT2)
            ]

            self.agent.bagpipe_port_attach(None, bagpipe_port10)
            self.agent.bgpvpn_port_attach(None, bgpvpn_port10)

            self.agent.bagpipe_port_attach(None, bagpipe_port20)
            self.agent.bgpvpn_port_attach(None, bgpvpn_port20)

            send_attach_fn.assert_has_calls(expected_calls)

            # Verify attachments list consistency
            for network_id in [NETWORK1['id'], NETWORK2['id']]:
                self._check_network_attachments(network_id,
                                                1,
                                                BAGPIPE_NOTIFIERS,
                                                [EVPN])

    def test_multiple_attach_multiple_ports_different_vpns(self):
        bagpipe_port10 = DummyPort(NETWORK1, PORT10,
                                   evpn=BAGPIPE_EVPN_RT1).__dict__
        bgpvpn_port10 = DummyPort(NETWORK1, PORT10,
                                  ipvpn=BGPVPN_IPVPN_RT100).__dict__

        bagpipe_port20 = DummyPort(NETWORK2, PORT20,
                                   evpn=BAGPIPE_EVPN_RT2).__dict__
        bgpvpn_port20 = DummyPort(NETWORK2, PORT20,
                                  ipvpn=BGPVPN_IPVPN_RT200).__dict__

        with mock.patch.object(self.agent,
                               'send_attach_local_port') as send_attach_fn:
            expected_calls = [
                self._mock_send_expected_call(EVPN,
                                              bagpipe_port10,
                                              self.DUMMY_VIF10),
                self._mock_send_expected_call(EVPN,
                                              bagpipe_port10,
                                              self.DUMMY_VIF10),
                self._mock_send_expected_call(IPVPN,
                                              bgpvpn_port10,
                                              self.DUMMY_VIF10,
                                              evpn2ipvpn=True),
                self._mock_send_expected_call(EVPN,
                                              bagpipe_port20,
                                              self.DUMMY_VIF20),
                self._mock_send_expected_call(EVPN,
                                              bagpipe_port20,
                                              self.DUMMY_VIF20),
                self._mock_send_expected_call(IPVPN,
                                              bgpvpn_port20,
                                              self.DUMMY_VIF20,
                                              evpn2ipvpn=True)
            ]

            self.agent.bagpipe_port_attach(None, bagpipe_port10)
            self.agent.bgpvpn_port_attach(None, bgpvpn_port10)

            self.agent.bagpipe_port_attach(None, bagpipe_port20)
            self.agent.bgpvpn_port_attach(None, bgpvpn_port20)

            send_attach_fn.assert_has_calls(expected_calls)

            # Verify attachments list consistency
            for network_id in [NETWORK1['id'], NETWORK2['id']]:
                self._check_network_attachments(network_id,
                                                1,
                                                [BAGPIPE_NOTIFIER],
                                                [EVPN])
                self._check_network_attachments(network_id,
                                                1,
                                                [BGPVPN_NOTIFIER],
                                                [IPVPN])

    def test_multiple_detach_single_port_evpns1(self):
        bagpipe_port10 = DummyPort(NETWORK1, PORT10,
                                   evpn=BAGPIPE_EVPN_RT1).__dict__
        bgpvpn_port10 = DummyPort(NETWORK1, PORT10,
                                  evpn=BGPVPN_EVPN_RT10).__dict__
        dummy_detach10 = dict(id=PORT10['id'], network_id=NETWORK1['id'])

        with mock.patch.object(self.agent,
                               'send_detach_local_port') as send_detach_fn:
            expected_calls = [
                self._mock_send_expected_call(EVPN,
                                              bgpvpn_port10,
                                              self.DUMMY_VIF10,
                                              others_rts=BAGPIPE_EVPN_RT1),
                self._mock_send_expected_call(EVPN,
                                              bagpipe_port10,
                                              self.DUMMY_VIF10)
            ]

            self.agent.bagpipe_port_attach(None, bagpipe_port10)
            self.agent.bgpvpn_port_attach(None, bgpvpn_port10)

            self.agent.bgpvpn_port_detach(None, dummy_detach10)

            # Verify attachments list consistency
            self._check_network_attachments(NETWORK1['id'],
                                            1,
                                            [BAGPIPE_NOTIFIER],
                                            [EVPN])

            self.agent.bagpipe_port_detach(None, dummy_detach10)

            send_detach_fn.assert_has_calls(expected_calls)

            # Verify attachments list consistency
            self._check_network_attachments(NETWORK1['id'], 0)

            self.assertEqual(0, len(self.agent.reg_attachments),
                             "Registered attachments list must be empty: %s" %
                             self.agent.reg_attachments)

    def test_multiple_detach_single_port_evpns2(self):
        bagpipe_port10 = DummyPort(NETWORK1, PORT10,
                                   evpn=BAGPIPE_EVPN_RT1).__dict__
        bgpvpn_port10 = DummyPort(NETWORK1, PORT10,
                                  evpn=BGPVPN_EVPN_RT10).__dict__
        dummy_detach10 = dict(id=PORT10['id'], network_id=NETWORK1['id'])

        with mock.patch.object(
            self.agent,
            'send_detach_local_port',
            side_effect=[None, BaGPipeBGPException(reason="Port not plugged")]
        ) as send_detach_fn:
            expected_calls = [
                self._mock_send_expected_call(EVPN,
                                              bagpipe_port10,
                                              self.DUMMY_VIF10,
                                              others_rts=BGPVPN_EVPN_RT10,
                                              others_last=True),
                self._mock_send_expected_call(EVPN,
                                              bgpvpn_port10,
                                              self.DUMMY_VIF10)
            ]

            self.agent.bagpipe_port_attach(None, bagpipe_port10)
            self.agent.bgpvpn_port_attach(None, bgpvpn_port10)

            self.agent.bagpipe_port_detach(None, dummy_detach10)
            # Verify attachments list consistency
            self._check_network_attachments(NETWORK1['id'],
                                            1,
                                            [BGPVPN_NOTIFIER],
                                            [EVPN])

            self.agent.bgpvpn_port_detach(None, dummy_detach10)

            send_detach_fn.assert_has_calls(expected_calls)

            # Verify attachments list consistency
            self._check_network_attachments(NETWORK1['id'], 0)

            self.assertEqual(0, len(self.agent.reg_attachments),
                             "Registered attachments list must be empty: %s" %
                             self.agent.reg_attachments)

    def test_multiple_detach_single_port_different_vpns(self):
        bagpipe_port10 = DummyPort(NETWORK1, PORT10,
                                   evpn=BAGPIPE_EVPN_RT1).__dict__
        bgpvpn_port10 = DummyPort(NETWORK1, PORT10,
                                  ipvpn=BGPVPN_IPVPN_RT100).__dict__
        dummy_detach10 = dict(id=PORT10['id'], network_id=NETWORK1['id'])

        with mock.patch.object(self.agent,
                               'send_detach_local_port') as send_detach_fn:
            expected_calls = [
                self._mock_send_expected_call(IPVPN,
                                              bgpvpn_port10,
                                              self.DUMMY_VIF10,
                                              evpn2ipvpn=True),
                self._mock_send_expected_call(EVPN,
                                              bagpipe_port10,
                                              self.DUMMY_VIF10)
            ]

            self.agent.bgpvpn_port_attach(None, bgpvpn_port10)
            self.agent.bagpipe_port_attach(None, bagpipe_port10)

            self.agent.bgpvpn_port_detach(None, dummy_detach10)
            self.agent.bagpipe_port_detach(None, dummy_detach10)

            send_detach_fn.assert_has_calls(expected_calls)

            # Verify attachments list consistency
            self._check_network_attachments(NETWORK1['id'], 0)

            self.assertEqual(0, len(self.agent.reg_attachments),
                             "Registered attachments list must be empty: %s" %
                             self.agent.reg_attachments)

    def test_multiple_detach_multiple_ports_same_evpns(self):
        bagpipe_port10 = DummyPort(NETWORK1, PORT10,
                                   evpn=BAGPIPE_EVPN_RT1).__dict__
        bgpvpn_port10 = DummyPort(NETWORK1, PORT10,
                                  evpn=BGPVPN_EVPN_RT10).__dict__
        dummy_detach10 = dict(id=PORT10['id'], network_id=NETWORK1['id'])

        bagpipe_port11 = DummyPort(NETWORK1, PORT11,
                                   evpn=BAGPIPE_EVPN_RT1).__dict__
        bgpvpn_port11 = DummyPort(NETWORK1, PORT11,
                                  evpn=BGPVPN_EVPN_RT10).__dict__
        dummy_detach11 = dict(id=PORT11['id'], network_id=NETWORK1['id'])

        with mock.patch.object(self.agent,
                               'send_detach_local_port') as send_detach_fn:
            expected_calls = [
                self._mock_send_expected_call(EVPN,
                                              bgpvpn_port10,
                                              self.DUMMY_VIF10,
                                              others_rts=BAGPIPE_EVPN_RT1),
                self._mock_send_expected_call(EVPN,
                                              bagpipe_port10,
                                              self.DUMMY_VIF10),
                self._mock_send_expected_call(EVPN,
                                              bgpvpn_port11,
                                              self.DUMMY_VIF11,
                                              others_rts=BAGPIPE_EVPN_RT1),
                self._mock_send_expected_call(EVPN,
                                              bagpipe_port11,
                                              self.DUMMY_VIF11)
            ]

            self.agent.bagpipe_port_attach(None, bagpipe_port10)
            self.agent.bgpvpn_port_attach(None, bgpvpn_port10)

            self.agent.bagpipe_port_attach(None, bagpipe_port11)
            self.agent.bgpvpn_port_attach(None, bgpvpn_port11)

            self.agent.bgpvpn_port_detach(None, dummy_detach10)
            self.agent.bagpipe_port_detach(None, dummy_detach10)

            self.agent.bgpvpn_port_detach(None, dummy_detach11)
            self.agent.bagpipe_port_detach(None, dummy_detach11)

            send_detach_fn.assert_has_calls(expected_calls)

            # Verify attachments list consistency
            self._check_network_attachments(NETWORK1['id'], 0)

            self.assertEqual(0, len(self.agent.reg_attachments),
                             "Registered attachments list must be empty: %s" %
                             self.agent.reg_attachments)

    def test_multiple_detach_multiple_ports_different_vpns(self):
        bagpipe_port10 = DummyPort(NETWORK1, PORT10,
                                   evpn=BAGPIPE_EVPN_RT1).__dict__
        bgpvpn_port10 = DummyPort(NETWORK1, PORT10,
                                  ipvpn=BGPVPN_IPVPN_RT100).__dict__
        dummy_detach10 = dict(id=PORT10['id'], network_id=NETWORK1['id'])

        bagpipe_port20 = DummyPort(NETWORK2, PORT20,
                                   evpn=BAGPIPE_EVPN_RT2).__dict__
        bgpvpn_port20 = DummyPort(NETWORK2, PORT20,
                                  ipvpn=BGPVPN_IPVPN_RT200).__dict__
        dummy_detach20 = dict(id=PORT20['id'], network_id=NETWORK2['id'])

        with mock.patch.object(
            self.agent,
            'send_detach_local_port',
            side_effect=[None, None,
                         BaGPipeBGPException(reason="Port not plugged"),
                         None, None,
                         BaGPipeBGPException(reason="Port not plugged")]
        ) as send_detach_fn:
            expected_calls = [
                self._mock_send_expected_call(IPVPN,
                                              bgpvpn_port10,
                                              self.DUMMY_VIF10,
                                              evpn2ipvpn=True),
                self._mock_send_expected_call(EVPN,
                                              bagpipe_port10,
                                              self.DUMMY_VIF10),
                self._mock_send_expected_call(EVPN,
                                              bagpipe_port10,
                                              self.DUMMY_VIF10),
                self._mock_send_expected_call(IPVPN,
                                              bgpvpn_port20,
                                              self.DUMMY_VIF20,
                                              evpn2ipvpn=True),
                self._mock_send_expected_call(EVPN,
                                              bagpipe_port20,
                                              self.DUMMY_VIF20),
                self._mock_send_expected_call(EVPN,
                                              bagpipe_port20,
                                              self.DUMMY_VIF20)
            ]

            self.agent.bagpipe_port_attach(None, bagpipe_port10)
            self.agent.bgpvpn_port_attach(None, bgpvpn_port10)

            self.agent.bagpipe_port_attach(None, bagpipe_port20)
            self.agent.bgpvpn_port_attach(None, bgpvpn_port20)

            self.agent.bgpvpn_port_detach(None, dummy_detach10)
            self.agent.bagpipe_port_detach(None, dummy_detach10)

            self.agent.bgpvpn_port_detach(None, dummy_detach20)
            self.agent.bagpipe_port_detach(None, dummy_detach20)

            send_detach_fn.assert_has_calls(expected_calls)

            # Verify attachments list consistency
            for network_id in [NETWORK1['id'], NETWORK2['id']]:
                self._check_network_attachments(network_id, 0)

            self.assertEqual(0, len(self.agent.reg_attachments),
                             "Registered attachments list must be empty: %s" %
                             self.agent.reg_attachments)


class TestBaGPipeBGPAgentLinuxBridge(base.BaseTestCase,
                                     TestBaGPipeBGPAgentMixin):

    def setUp(self):
        cfg.CONF.set_override('enable_vxlan', False, 'VXLAN')
        super(TestBaGPipeBGPAgentLinuxBridge, self).setUp()

        self.bridge_mappings = {}
        self.interface_mappings = {}

        self.agent = BaGPipeBGPAgent(n_const.AGENT_TYPE_LINUXBRIDGE,
                                     mock.Mock())

    def _get_expected_local_port(self, network_id, port_id, vif_name):
        local_port = dict(
            linuxif=LinuxBridgeManager.get_tap_device_name(port_id)
        )
        linuxbr = LinuxBridgeManager.get_bridge_name(network_id)

        return local_port, linuxbr


class TestBaGPipeBGPAgentOVS(ovs_test_base.OVSOFCtlTestBase,
                             TestBaGPipeBGPAgentMixin):

    DUMMY_VIF10 = DummyVif(10, 'VIF10')
    DUMMY_VIF11 = DummyVif(11, 'VIF11')
    DUMMY_VIF20 = DummyVif(20, 'VIF20')
    DUMMY_VIF21 = DummyVif(21, 'VIF21')

    def setUp(self):
        super(TestBaGPipeBGPAgentOVS, self).setUp()

        self.mock_int_br = mock.Mock()
        self.mock_int_br.add_patch_port = mock.Mock()
        self.mock_int_br.add_patch_port.side_effect = PATCH_INT_OFPORTS
        self.mock_int_br.get_vif_port_by_id = mock.Mock()

        self.mock_tun_br = mock.Mock()
        self.mock_tun_br.add_patch_port = mock.Mock()
        self.mock_tun_br.add_patch_port.side_effect = PATCH_TUN_OFPORTS
        self.mock_tun_br.get_port_ofport = mock.Mock()

        with mock.patch('neutron.agent.common.ovs_lib.OVSBridge.'
                        'bridge_exists', return_value=True), \
                mock.patch('neutron.agent.common.ovs_lib.OVSBridge.'
                           'add_patch_port', side_effect=PATCH_MPLS_OFPORTS):
            self.agent = BaGPipeBGPAgent(n_const.AGENT_TYPE_OVS,
                                         mock.Mock(),
                                         int_br=self.mock_int_br,
                                         tun_br=self.mock_tun_br)

        self.vlan_manager = vlanmanager.LocalVlanManager()
        for net_id, vlan in LOCAL_VLAN_MAP.items():
            try:
                self.vlan_manager.add(net_id, vlan, None, None, None)
            except vlanmanager.MappingAlreadyExists:
                pass

    def _get_expected_local_port(self, network_id, port_id, vif_name):
        vlan = self.vlan_manager.get(network_id).vlan
        local_port = dict(
            linuxif="patch2tun:%s" % vlan,
            ovs=dict(plugged=True,
                     port_number=PATCH_MPLS_FROM_TUN_OFPORT,
                     to_vm_port_number=PATCH_MPLS_TO_TUN_OFPORT,
                     vlan=vlan)
        )

        return local_port, None

    def test_bagpipe_attach_single_port(self):
        with mock.patch.object(self.agent.int_br, 'get_vif_port_by_id',
                               return_value=self.DUMMY_VIF10):
            super(TestBaGPipeBGPAgentOVS,
                  self).test_bagpipe_attach_single_port()

    def test_bagpipe_attach_same_port_different_route_target(self):
        with mock.patch.object(self.agent.int_br, 'get_vif_port_by_id',
                               return_value=self.DUMMY_VIF10):
            super(TestBaGPipeBGPAgentOVS,
                  self).test_bagpipe_attach_same_port_different_route_target()

    def test_bagpipe_attach_multiple_ports_same_network(self):
        with mock.patch.object(self.agent.int_br, 'get_vif_port_by_id',
                               side_effect=[self.DUMMY_VIF10,
                                            self.DUMMY_VIF11]):
            super(TestBaGPipeBGPAgentOVS,
                  self).test_bagpipe_attach_multiple_ports_same_network()

    def test_bagpipe_attach_multiple_ports_different_networks(self):
        with mock.patch.object(self.agent.int_br, 'get_vif_port_by_id',
                               side_effect=[self.DUMMY_VIF10,
                                            self.DUMMY_VIF20]):
            super(TestBaGPipeBGPAgentOVS,
                  self).test_bagpipe_attach_multiple_ports_different_networks()

    def test_bagpipe_detach_single_port(self):
        with mock.patch.object(self.agent.int_br, 'get_vif_port_by_id',
                               return_value=self.DUMMY_VIF10):
            super(TestBaGPipeBGPAgentOVS,
                  self).test_bagpipe_detach_single_port()

    def test_bagpipe_detach_multiple_ports_same_network(self):
        with mock.patch.object(self.agent.int_br, 'get_vif_port_by_id',
                               side_effect=[self.DUMMY_VIF10,
                                            self.DUMMY_VIF11]):
            super(TestBaGPipeBGPAgentOVS,
                  self).test_bagpipe_detach_multiple_ports_same_network()

    def test_bagpipe_detach_multiple_ports_different_networks(self):
        with mock.patch.object(self.agent.int_br, 'get_vif_port_by_id',
                               side_effect=[self.DUMMY_VIF10,
                                            self.DUMMY_VIF11,
                                            self.DUMMY_VIF20,
                                            self.DUMMY_VIF21]):
            super(TestBaGPipeBGPAgentOVS,
                  self).test_bagpipe_detach_multiple_ports_different_networks()

    # ----------------------------
    # BGP VPN RPC notifier tests |
    # ----------------------------
    @mock.patch('networking_bagpipe.agent.bagpipe_bgp_agent'
                '.BaGPipeBGPAgent._enable_gw_redirect',
                autospec=True)
    def test_update_bgpvpn_already_plugged_ports(self, gw_redir_fn):
        with mock.patch.object(self.agent.int_br, 'get_vif_port_by_id',
                               side_effect=[self.DUMMY_VIF10,
                                            self.DUMMY_VIF11]):
            super(TestBaGPipeBGPAgentOVS,
                  self).test_update_bgpvpn_already_plugged_ports()
            self.assertEqual(2, gw_redir_fn.call_count)

    def test_update_bgpvpn_same_vpn_types(self):
        with mock.patch.object(self.agent.int_br, 'get_vif_port_by_id',
                               return_value=self.DUMMY_VIF10):
            super(TestBaGPipeBGPAgentOVS,
                  self).test_update_bgpvpn_same_vpn_types()

    @mock.patch('networking_bagpipe.agent.bagpipe_bgp_agent'
                '.BaGPipeBGPAgent._enable_gw_redirect',
                autospec=True)
    def test_update_bgpvpn_different_vpn_types(self, gw_redir_fn):
        with mock.patch.object(self.agent.int_br, 'get_vif_port_by_id',
                               return_value=self.DUMMY_VIF10):
            super(TestBaGPipeBGPAgentOVS,
                  self).test_update_bgpvpn_different_vpn_types()
            self.assertEqual(1, gw_redir_fn.call_count)

    @mock.patch('networking_bagpipe.agent.bagpipe_bgp_agent'
                '.BaGPipeBGPAgent._enable_gw_redirect',
                autospec=True)
    def test_delete_bgpvpn_remaining_plugged_ports(self, gw_redir_fn):
        with mock.patch.object(self.agent.int_br, 'get_vif_port_by_id',
                               side_effect=[self.DUMMY_VIF10,
                                            self.DUMMY_VIF11]):
            super(TestBaGPipeBGPAgentOVS,
                  self).test_delete_bgpvpn_remaining_plugged_ports()

    @mock.patch('networking_bagpipe.agent.bagpipe_bgp_agent'
                '.BaGPipeBGPAgent._enable_gw_redirect',
                autospec=True)
    def test_delete_bgpvpn_had_plugged_ports(self, gw_redir_fn):
        with mock.patch.object(self.agent.int_br, 'get_vif_port_by_id',
                               side_effect=[self.DUMMY_VIF10,
                                            self.DUMMY_VIF11]):
            super(TestBaGPipeBGPAgentOVS,
                  self).test_delete_bgpvpn_had_plugged_ports()

            self.assertEqual(2, gw_redir_fn.call_count)

    @mock.patch('networking_bagpipe.agent.bagpipe_bgp_agent'
                '.BaGPipeBGPAgent._enable_gw_redirect',
                autospec=True)
    def _test_bgpvpn_attach_single_port(self, bgpvpn, network,
                                        gw_redir_fn):
        with mock.patch.object(self.agent.int_br, 'get_vif_port_by_id',
                               return_value=self.DUMMY_VIF10):
            super(TestBaGPipeBGPAgentOVS,
                  self)._test_bgpvpn_attach_single_port(bgpvpn, network)
            self.assertEqual(1 if (bgpvpn == IPVPN or
                                   bgpvpn == BGPVPN_L3) else 0,
                             gw_redir_fn.call_count)

    @mock.patch('networking_bagpipe.agent.bagpipe_bgp_agent'
                '.BaGPipeBGPAgent._enable_gw_redirect',
                autospec=True)
    def test_bgpvpn_attach_same_port_different_bgpvpn(self, gw_redir_fn):
        with mock.patch.object(self.agent.int_br, 'get_vif_port_by_id',
                               return_value=self.DUMMY_VIF10):
            super(TestBaGPipeBGPAgentOVS,
                  self).test_bgpvpn_attach_same_port_different_bgpvpn()
            self.assertEqual(1, gw_redir_fn.call_count)

    @mock.patch('networking_bagpipe.agent.bagpipe_bgp_agent'
                '.BaGPipeBGPAgent._enable_gw_redirect',
                autospec=True)
    def test_bgpvpn_attach_single_port_multiple_bgpvpns(self, gw_redir_fn):
        with mock.patch.object(self.agent.int_br, 'get_vif_port_by_id',
                               return_value=self.DUMMY_VIF10):
            super(TestBaGPipeBGPAgentOVS,
                  self).test_bgpvpn_attach_single_port_multiple_bgpvpns()
            self.assertEqual(1, gw_redir_fn.call_count)

    @mock.patch('networking_bagpipe.agent.bagpipe_bgp_agent'
                '.BaGPipeBGPAgent._enable_gw_redirect',
                autospec=True)
    def test_bgpvpn_attach_multiple_ports_same_bgpvpn(self, gw_redir_fn):
        with mock.patch.object(self.agent.int_br, 'get_vif_port_by_id',
                               side_effect=[self.DUMMY_VIF10,
                                            self.DUMMY_VIF11]):
            super(TestBaGPipeBGPAgentOVS,
                  self).test_bgpvpn_attach_multiple_ports_same_bgpvpn()
            self.assertEqual(2, gw_redir_fn.call_count)

    @mock.patch('networking_bagpipe.agent.bagpipe_bgp_agent'
                '.BaGPipeBGPAgent._enable_gw_redirect',
                autospec=True)
    def test_bgpvpn_attach_multiple_ports_different_bgpvpns(self,
                                                            gw_redir_fn):
        with mock.patch.object(self.agent.int_br, 'get_vif_port_by_id',
                               side_effect=[self.DUMMY_VIF10,
                                            self.DUMMY_VIF20]):
            super(TestBaGPipeBGPAgentOVS,
                  self).test_bgpvpn_attach_multiple_ports_different_bgpvpns()
            self.assertEqual(2, gw_redir_fn.call_count)

    def _test_bgpvpn_detach_single_port(self, bgpvpn, network):
        with mock.patch.object(self.agent.int_br, 'get_vif_port_by_id',
                               return_value=self.DUMMY_VIF10):
            super(TestBaGPipeBGPAgentOVS,
                  self)._test_bgpvpn_detach_single_port(bgpvpn, network)

    def test_bgpvpn_detach_single_port_multiple_bgpvpns(self):
        with mock.patch.object(self.agent.int_br, 'get_vif_port_by_id',
                               return_value=self.DUMMY_VIF10):
            super(TestBaGPipeBGPAgentOVS,
                  self).test_bgpvpn_detach_single_port_multiple_bgpvpns()

    def test_bgpvpn_detach_multiple_ports_same_bgpvpn(self):
        with mock.patch.object(self.agent.int_br, 'get_vif_port_by_id',
                               side_effect=[self.DUMMY_VIF10,
                                            self.DUMMY_VIF11]):
            super(TestBaGPipeBGPAgentOVS,
                  self).test_bgpvpn_detach_multiple_ports_same_bgpvpn()

    def test_bgpvpn_detach_multiple_ports_different_bgpvpns(self):
        with mock.patch.object(self.agent.int_br, 'get_vif_port_by_id',
                               side_effect=[self.DUMMY_VIF10,
                                            self.DUMMY_VIF20]):
            super(TestBaGPipeBGPAgentOVS,
                  self).test_bgpvpn_detach_multiple_ports_different_bgpvpns()

    # Test fallback and ARP gateway voodoo

    def test_fallback(self):
        GW_MAC = 'aa:bb:cc:dd:ee:ff'

        with mock.patch.object(self.agent.int_br, 'get_vif_port_by_id',
                               side_effect=[self.DUMMY_VIF10,
                                            self.DUMMY_VIF11]):
            super(TestBaGPipeBGPAgentOVS,
                  self).test_update_bgpvpn_already_plugged_ports()

            port10_with_gw_mac = DummyPort(NETWORK1, PORT10).__dict__
            port10_with_gw_mac.update({'gateway_mac': GW_MAC})
            port10_with_gw_mac.update(DummyBGPVPN(NETWORK1,
                                                  ipvpn=BGPVPN_IPVPN_RT100).
                                      __dict__)
            with mock.patch.object(self.agent, 'send_attach_local_port') as\
                    send_attach_fn:
                self.agent.bgpvpn_port_attach(None,
                                              copy.copy(port10_with_gw_mac))

                fallback = {'dst_mac': GW_MAC,
                            'ovs_port_number': PATCH_MPLS_TO_INT_OFPORT,
                            'src_mac': '00:00:5e:2a:10:00'}

                expected_calls = [
                    self._mock_send_expected_call(IPVPN,
                                                  port10_with_gw_mac,
                                                  self.DUMMY_VIF10,
                                                  fallback=fallback),
                ]

                send_attach_fn.assert_has_calls(expected_calls)

    def test_gateway_arp_voodoo(self):
        GW_MAC = 'aa:bb:cc:dd:ee:ff'

        with mock.patch.object(self.agent.int_br, 'get_vif_port_by_id',
                               side_effect=[self.DUMMY_VIF10,
                                            self.DUMMY_VIF11]), \
                mock.patch.object(self.agent.int_br,
                                  'add_flow') as add_flow, \
                mock.patch.object(self.agent.tun_br,
                                  'delete_flows') as tun_delete_flows,\
                mock.patch.object(self.agent.int_br,
                                  'delete_flows') as int_delete_flows:
            super(TestBaGPipeBGPAgentOVS,
                  self).test_update_bgpvpn_already_plugged_ports()

            port10_with_gw_mac = DummyPort(NETWORK1, PORT10).__dict__
            port10_with_gw_mac.update({'gateway_mac': GW_MAC})
            port10_with_gw_mac.update(DummyBGPVPN(NETWORK1,
                                                  ipvpn=BGPVPN_IPVPN_RT100).
                                      __dict__)

            self.agent.bgpvpn_port_attach(None, copy.copy(port10_with_gw_mac))

            self.assertEqual(2, add_flow.call_count)

            add_flow.assert_has_calls([
                mock.call(table=mock.ANY,
                          priority=2,
                          proto='arp',
                          arp_op=0x2,
                          dl_src=GW_MAC,
                          arp_sha=GW_MAC,
                          arp_spa='10.0.0.1',
                          actions="drop"),
                mock.call(table=mock.ANY,
                          priority=2,
                          proto='arp',
                          arp_op=0x01,
                          dl_src=GW_MAC,
                          arp_spa='10.0.0.1',
                          arp_sha=GW_MAC,
                          actions="load:0x0->NXM_OF_ARP_SPA[],NORMAL"
                          )
            ])

            self.agent.bgpvpn_port_detach(None, DummyPort(NETWORK1,
                                                          PORT10).__dict__)

            self.assertEqual(0, tun_delete_flows.call_count)
            self.assertEqual(0, int_delete_flows.call_count)

            self.agent.bgpvpn_port_detach(None, DummyPort(NETWORK1,
                                                          PORT11).__dict__)

            self.assertEqual(1, tun_delete_flows.call_count)
            self.assertEqual(1, int_delete_flows.call_count)

            tun_delete_flows.assert_has_calls([
                mock.call(table=mock.ANY,
                          proto='arp',
                          arp_op=0x01,
                          arp_tpa='10.0.0.1',
                          dl_vlan=mock.ANY,
                          )])
            int_delete_flows.assert_has_calls([
                mock.call(table=mock.ANY,
                          proto='arp',
                          dl_src=GW_MAC,
                          arp_sha=GW_MAC),
            ])

    def test_gateway_arp_voodoo_update_bgpvpn_after_plug(self):
        GW_MAC = 'aa:bb:cc:dd:ee:ff'

        with mock.patch.object(self.agent.int_br, 'get_vif_port_by_id',
                               side_effect=[self.DUMMY_VIF10,
                                            self.DUMMY_VIF11]), \
                mock.patch.object(self.agent.int_br,
                                  'add_flow') as add_flow:
            port10 = DummyPort(NETWORK1, PORT10).__dict__
            port10.update({'gateway_mac': GW_MAC})

            self.agent.bgpvpn_port_attach(None, copy.copy(port10))

            dummy_bgpvpn1 = DummyBGPVPN(NETWORK1,
                                        ipvpn=BGPVPN_IPVPN_RT100,
                                        gateway_mac=GW_MAC).__dict__

            with mock.patch.object(self.agent, 'send_attach_local_port') as\
                    send_attach_fn:

                self.agent.update_bgpvpn(None, dummy_bgpvpn1)

                self.assertEqual(2, add_flow.call_count)

                add_flow.assert_has_calls([
                    mock.call(table=mock.ANY,
                              priority=2,
                              proto='arp',
                              arp_op=0x2,
                              dl_src=GW_MAC,
                              arp_sha=GW_MAC,
                              arp_spa='10.0.0.1',
                              actions="drop"),
                    mock.call(table=mock.ANY,
                              priority=2,
                              proto='arp',
                              arp_op=0x01,
                              dl_src=GW_MAC,
                              arp_spa='10.0.0.1',
                              arp_sha=GW_MAC,
                              actions="load:0x0->NXM_OF_ARP_SPA[],NORMAL"
                              )
                ])

                fallback = {'dst_mac': GW_MAC,
                            'ovs_port_number': PATCH_MPLS_TO_INT_OFPORT,
                            'src_mac': '00:00:5e:2a:10:00'}

                expected_calls = [
                    self._mock_send_expected_call(
                        IPVPN,
                        port10,
                        self.DUMMY_VIF10,
                        others_rts=BGPVPN_IPVPN_RT100,
                        fallback=fallback),
                ]

                send_attach_fn.assert_has_calls(expected_calls)

                self.agent.delete_bgpvpn(None, dummy_bgpvpn1)

                add_flow.reset_mock()
                self.assertEqual(0, add_flow.call_count)

                self.agent.update_bgpvpn(None, dummy_bgpvpn1)

                self.assertEqual(2, add_flow.call_count)

    def test_gateway_plug_before_update(self):
        GW_MAC = 'aa:bb:cc:dd:ee:ff'

        with mock.patch.object(self.agent.int_br, 'get_vif_port_by_id',
                               side_effect=[self.DUMMY_VIF10]), \
                mock.patch.object(self.agent.int_br,
                                  'add_flow') as add_flow:
            port10 = DummyPort(NETWORK1, PORT10).__dict__

            self.agent.bgpvpn_port_attach(None, copy.copy(port10))

            dummy_bgpvpn1 = DummyBGPVPN(NETWORK1,
                                        ipvpn=BGPVPN_IPVPN_RT100,
                                        gateway_mac=GW_MAC).__dict__

            with mock.patch.object(self.agent, 'send_attach_local_port') as\
                    send_attach_fn:

                self.agent.update_bgpvpn(None, dummy_bgpvpn1)

                self.assertEqual(2, add_flow.call_count)

                add_flow.assert_has_calls([
                    mock.call(table=mock.ANY,
                              priority=2,
                              proto='arp',
                              arp_op=0x2,
                              dl_src=GW_MAC,
                              arp_sha=GW_MAC,
                              arp_spa='10.0.0.1',
                              actions="drop"),
                    mock.call(table=mock.ANY,
                              priority=2,
                              proto='arp',
                              arp_op=0x01,
                              dl_src=GW_MAC,
                              arp_spa='10.0.0.1',
                              arp_sha=GW_MAC,
                              actions="load:0x0->NXM_OF_ARP_SPA[],NORMAL"
                              )
                ])

                fallback = {'dst_mac': GW_MAC,
                            'ovs_port_number': PATCH_MPLS_TO_INT_OFPORT,
                            'src_mac': '00:00:5e:2a:10:00'}

                expected_calls = [
                    self._mock_send_expected_call(
                        IPVPN,
                        port10,
                        self.DUMMY_VIF10,
                        others_rts=BGPVPN_IPVPN_RT100,
                        fallback=fallback),
                ]

                send_attach_fn.assert_has_calls(expected_calls)

                self.agent.delete_bgpvpn(None, dummy_bgpvpn1)

                add_flow.reset_mock()
                self.assertEqual(0, add_flow.call_count)

                self.agent.update_bgpvpn(None, dummy_bgpvpn1)

                self.assertEqual(2, add_flow.call_count)

    def test_evpn_no_gateway_arp_voodoo(self):
        GW_MAC = 'aa:bb:cc:dd:ee:ff'

        with mock.patch.object(self.agent.int_br, 'get_vif_port_by_id',
                               side_effect=[self.DUMMY_VIF10,
                                            self.DUMMY_VIF11]), \
                mock.patch.object(self.agent.int_br,
                                  'add_flow') as add_flow, \
                mock.patch.object(self.agent.int_br,
                                  'delete_flows') as delete_flows:

            port10_with_gw_mac = DummyPort(NETWORK1, PORT10).__dict__
            port10_with_gw_mac.update({'gateway_mac': GW_MAC})
            port10_with_gw_mac.update(DummyBGPVPN(NETWORK1,
                                                  evpn=BGPVPN_EVPN_RT10).
                                      __dict__)

            self.agent.bgpvpn_port_attach(None, copy.copy(port10_with_gw_mac))

            self.assertEqual(0, add_flow.call_count)

            self.agent.bgpvpn_port_detach(None, DummyPort(NETWORK1,
                                                          PORT10).__dict__)
            self.assertEqual(0, delete_flows.call_count)

            self.agent.bgpvpn_port_detach(None, DummyPort(NETWORK1,
                                                          PORT11).__dict__)

            self.assertEqual(0, delete_flows.call_count)

    def test_gateway_arp_voodoo_ovs_restart(self):
        GW_MAC = 'aa:bb:cc:dd:ee:ff'

        with mock.patch.object(self.agent.int_br, 'get_vif_port_by_id',
                               side_effect=[self.DUMMY_VIF10,
                                            self.DUMMY_VIF11]), \
                mock.patch.object(self.agent.int_br,
                                  'add_flow') as add_flow:
            port10 = DummyPort(NETWORK1, PORT10).__dict__
            port10.update({'gateway_mac': GW_MAC})

            self.agent.bgpvpn_port_attach(None, copy.copy(port10))

            dummy_bgpvpn1 = DummyBGPVPN(NETWORK1,
                                        ipvpn=BGPVPN_IPVPN_RT100,
                                        gateway_mac=GW_MAC).__dict__

            self.agent.update_bgpvpn(None, dummy_bgpvpn1)

            self.assertEqual(2, add_flow.call_count)

            expected_calls = [
                mock.call(table=mock.ANY,
                          priority=2,
                          proto='arp',
                          arp_op=0x2,
                          dl_src=GW_MAC,
                          arp_sha=GW_MAC,
                          arp_spa='10.0.0.1',
                          actions="drop"),
                mock.call(table=mock.ANY,
                          priority=2,
                          proto='arp',
                          arp_op=0x01,
                          dl_src=GW_MAC,
                          arp_spa='10.0.0.1',
                          arp_sha=GW_MAC,
                          actions="load:0x0->NXM_OF_ARP_SPA[],NORMAL"
                          )
            ]
            add_flow.assert_has_calls(expected_calls)

            add_flow.reset_mock()

            self.agent.ovs_restarted_bgpvpn()

            add_flow.assert_has_calls(expected_calls)

    # -------------------------------------------
    # Multiple simultaneous RPC notifiers tests |
    # -------------------------------------------

    def test_multiple_attach_single_port_evpns(self):
        with mock.patch.object(self.agent.int_br, 'get_vif_port_by_id',
                               return_value=self.DUMMY_VIF10):
            super(TestBaGPipeBGPAgentOVS,
                  self).test_multiple_attach_single_port_evpns()

    def test_multiple_attach_single_port_different_vpns(self):
        with mock.patch.object(self.agent.int_br, 'get_vif_port_by_id',
                               return_value=self.DUMMY_VIF10):
            super(TestBaGPipeBGPAgentOVS,
                  self).test_multiple_attach_single_port_different_vpns()

    def test_multiple_attach_multiple_ports_same_evpns(self):
        with mock.patch.object(self.agent.int_br, 'get_vif_port_by_id',
                               side_effect=[self.DUMMY_VIF10,
                                            self.DUMMY_VIF11]):
            super(TestBaGPipeBGPAgentOVS,
                  self).test_multiple_attach_multiple_ports_same_evpns()

    def test_multiple_attach_multiple_ports_different_evpns(self):
        with mock.patch.object(self.agent.int_br, 'get_vif_port_by_id',
                               side_effect=[self.DUMMY_VIF10,
                                            self.DUMMY_VIF20]):
            super(TestBaGPipeBGPAgentOVS,
                  self).test_multiple_attach_multiple_ports_different_evpns()

    def test_multiple_attach_multiple_ports_different_vpns(self):
        with mock.patch.object(self.agent.int_br, 'get_vif_port_by_id',
                               side_effect=[self.DUMMY_VIF10,
                                            self.DUMMY_VIF20]):
            super(TestBaGPipeBGPAgentOVS,
                  self).test_multiple_attach_multiple_ports_different_vpns()

    def test_multiple_detach_single_port_evpns1(self):
        with mock.patch.object(self.agent.int_br, 'get_vif_port_by_id',
                               return_value=self.DUMMY_VIF10):
            super(TestBaGPipeBGPAgentOVS,
                  self).test_multiple_detach_single_port_evpns1()

    def test_multiple_detach_single_port_evpns2(self):
        with mock.patch.object(self.agent.int_br, 'get_vif_port_by_id',
                               return_value=self.DUMMY_VIF10):
            super(TestBaGPipeBGPAgentOVS,
                  self).test_multiple_detach_single_port_evpns2()

    def test_multiple_detach_single_port_different_vpns(self):
        with mock.patch.object(self.agent.int_br, 'get_vif_port_by_id',
                               return_value=self.DUMMY_VIF10):
            super(TestBaGPipeBGPAgentOVS,
                  self).test_multiple_detach_single_port_different_vpns()

    def test_multiple_detach_multiple_ports_same_evpns(self):
        with mock.patch.object(self.agent.int_br, 'get_vif_port_by_id',
                               side_effect=[self.DUMMY_VIF10,
                                            self.DUMMY_VIF11]):
            super(TestBaGPipeBGPAgentOVS,
                  self).test_multiple_detach_multiple_ports_same_evpns()

    def test_multiple_detach_multiple_ports_different_vpns(self):
        with mock.patch.object(self.agent.int_br, 'get_vif_port_by_id',
                               side_effect=[self.DUMMY_VIF10,
                                            self.DUMMY_VIF20]):
            super(TestBaGPipeBGPAgentOVS,
                  self).test_multiple_detach_multiple_ports_different_vpns()



# vim: tabstop=4 shiftwidth=4 softtabstop=4
# encoding: utf-8

# Copyright 2014 Orange
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from oslo_config import cfg
# from oslo_config import types as oslo_types

cli_opts = [
    cfg.StrOpt("action", positional=True, default='unset',
               choices=('start', 'stop', 'unset'),
               help=("(deprecated, can be omitted)"),
               deprecated_for_removal=True
               )
]

bgp_opts = [
    cfg.IPOpt('local_address', required=True,
              help="IP address used for BGP peerings"),
    cfg.ListOpt('peers', default=[],  # NOTE(tmorin): use item_type=
                                      # oslo_types.IPAddress
                help="IP addresses of BGP peers"),
    cfg.IntOpt('my_as', min=1, max=2**16-1, required=True,
               help="Our BGP Autonomous System"),
    cfg.BoolOpt('enable_rtc', default=True,
                help="Enable RT Constraint (RFC4684)"),
    cfg.PortOpt('bgp_port', default=179,
                help="TCP port of connections to BGP peers")
]


def register():
    cfg.CONF.register_cli_opts(cli_opts)
    cfg.CONF.register_opts(bgp_opts, "BGP")



# Copyright (c) 2015 Orange.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
import six

from oslo_config import cfg

from oslo_log import log

import sqlalchemy as sa
from sqlalchemy.orm import exc as sa_exc

from networking_bagpipe._i18n import _

from neutron.db import api as db_api
from neutron.plugins.ml2 import driver_api as api
from neutron.plugins.ml2.drivers import helpers

from neutron_lib.db import model_base
from neutron_lib import exceptions as exc

LOG = log.getLogger(__name__)

MIN_RT_NN = 1
MAX_RT_NN = 2 ** 16 - 1

route_target_opts = [
    cfg.IntOpt('rt_asn', default=64512,
               help=_("Route Target Autonomous System number.")),
    cfg.ListOpt('rt_nn_ranges',
                default=[],
                help=_("Comma-separated list of <rt_nn_min>:<rt_nn_max> tuples"
                       " enumerating ranges of Route Target number that are "
                       "available for tenant network allocation")),
]

cfg.CONF.register_opts(route_target_opts, "ml2_type_route_target")

# TODO(tmorin): find a clean way to not collide in the segment type namespace
TYPE_ROUTE_TARGET = 'route_target'


class NetworkRouteTargetRangeError(exc.NeutronException):
    message = _("Invalid network Route Target number range: '%(rt_nn_range)s' "
                "- '%(error)s'")

    def __init__(self, **kwargs):
        # Convert rt_nn_range tuple to 'start:end' format for display
        if isinstance(kwargs['rt_nn_range'], tuple):
            kwargs['rt_nn_range'] = "%d:%d" % kwargs['rt_nn_range']
        super(NetworkRouteTargetRangeError, self).__init__(**kwargs)


def is_valid_route_target_number(rt_nn):
    return MIN_RT_NN <= rt_nn <= MAX_RT_NN


def verify_route_target_number_range(rt_nn_range):
    """Raise an exception for invalid tags or malformed range."""
    for rt_nn in rt_nn_range:
        if not is_valid_route_target_number(rt_nn):
            raise NetworkRouteTargetRangeError(
                rt_nn_range=rt_nn_range,
                error="%s is not a valid Route Target number" % rt_nn)
    if rt_nn_range[1] < rt_nn_range[0]:
        raise NetworkRouteTargetRangeError(
            rt_nn_range=rt_nn_range,
            error=_("End of Route Target number range is less than start of "
                    "Route Target number range"))


class RouteTargetInUse(exc.InUse):
    message = _("Unable to create the network. "
                "The route target %(rt_nn)s is in use.")


class RouteTargetAllocation(model_base.BASEV2):

    __tablename__ = 'ml2_route_target_allocations'

    rt_nn = sa.Column(sa.Integer, nullable=False, primary_key=True,
                      autoincrement=False)
    allocated = sa.Column(sa.Boolean, nullable=False, default=False)


class RouteTargetTypeDriver(helpers.SegmentTypeDriver):

    def __init__(self):
        super(RouteTargetTypeDriver, self).__init__(RouteTargetAllocation)

    def get_type(self):
        return TYPE_ROUTE_TARGET

    def initialize(self):
        self.rt_nn_ranges = []
        self._parse_route_target_number_ranges(
            cfg.CONF.ml2_type_route_target.rt_nn_ranges, self.rt_nn_ranges)
        self._sync_route_target_allocations()

    def _parse_route_target_number_ranges(self, rt_nn_ranges, current_range):
        for entry in rt_nn_ranges:
            entry = entry.strip()
            try:
                rt_nn_min, rt_nn_max = entry.split(':')
                rt_nn_min = rt_nn_min.strip()
                rt_nn_max = rt_nn_max.strip()
                rt_nn_range = int(rt_nn_min), int(rt_nn_max)
            except ValueError as ex:
                raise NetworkRouteTargetRangeError(rt_nn_range=entry, error=ex)
            verify_route_target_number_range(rt_nn_range)
            current_range.append(rt_nn_range)
        LOG.info("Route Target number ranges: %(range)s",
                 {'range': current_range})

    def is_partial_segment(self, segment):
        return segment.get(api.SEGMENTATION_ID) is None

    def validate_provider_segment(self, segment):
        physical_network = segment.get(api.PHYSICAL_NETWORK)
        if physical_network:
            msg = _("provider:physical_network specified for %s "
                    "network") % segment.get(api.NETWORK_TYPE)
            raise exc.InvalidInput(error_message=msg)

        for key, value in six.iteritems(segment):
            if value and key not in [api.NETWORK_TYPE,
                                     api.SEGMENTATION_ID]:
                msg = _("%s prohibited for Route Target provider \
                        network") % key
                raise exc.InvalidInput(error_message=msg)

    def reserve_provider_segment(self, context, segment):
        filters = {}
        rt_nn = segment.get(api.SEGMENTATION_ID)
        if rt_nn is not None:
                filters['rt_nn'] = rt_nn

        if self.is_partial_segment(segment):
            alloc = self.allocate_partially_specified_segment(
                context, **filters)
            if not alloc:
                raise exc.NoNetworkAvailable()
        else:
            alloc = self.allocate_fully_specified_segment(
                context, **filters)
            if not alloc:
                raise RouteTargetInUse(**filters)

        return {api.NETWORK_TYPE: TYPE_ROUTE_TARGET,
                api.PHYSICAL_NETWORK: None,
                api.SEGMENTATION_ID: alloc.rt_nn}

    def allocate_tenant_segment(self, context):
        alloc = self.allocate_partially_specified_segment(context)
        if not alloc:
            return
        return {api.NETWORK_TYPE: TYPE_ROUTE_TARGET,
                api.PHYSICAL_NETWORK: None,
                api.SEGMENTATION_ID: alloc.rt_nn}

    def release_segment(self, context, segment):
        rt_nn = segment[api.SEGMENTATION_ID]
        with context.session.begin(subtransactions=True):
            try:
                alloc = (context.session.query(RouteTargetAllocation).
                         filter_by(rt_nn=rt_nn).
                         with_lockmode('update').
                         one())
                alloc.allocated = False
                for lo, hi in self.rt_nn_ranges:
                    if lo <= rt_nn <= hi:
                        LOG.debug("Releasing Route Target number %s to "
                                  "pool", rt_nn)
                        break
                else:
                    context.session.delete(alloc)
                    LOG.debug("Releasing Route Target number %s outside "
                              "pool", rt_nn)
            except sa_exc.NoResultFound:
                LOG.warning("Route Target number %(rt_nn)s not found",
                            {'rt_nn': rt_nn})

    def _sync_route_target_allocations(self):
        """Synchronize route_target_allocations table with configured ranges"""

        # Determine current configured allocatable route targets
        rt_nns = set()
        for rt_nn_range in self.rt_nn_ranges:
            rt_nn_min, rt_nn_max = rt_nn_range
            if rt_nn_max + 1 - rt_nn_min > MAX_RT_NN:
                LOG.error("Skipping unreasonable route target range "
                          "%(rt_nn_min)s:%(rt_nn_max)s",
                          {'rt_nn_min': rt_nn_min, 'rt_nn_max': rt_nn_max})
            else:
                rt_nns |= set(six.moves.range(rt_nn_min, rt_nn_max + 1))

        session = db_api.get_writer_session()
        with session.begin(subtransactions=True):
            # Remove from table unallocated route target not currently
            # allocatable
            allocs = session.query(RouteTargetAllocation
                                   ).with_lockmode("update")
            for alloc in allocs:
                try:
                    # See if route target is allocatable
                    rt_nns.remove(alloc.rt_nn)
                except KeyError:
                    # Route target not allocatable, so check if its allocated
                    if not alloc.allocated:
                        # Not allocated, so remove it from table
                        LOG.debug("Removing route target %s from pool",
                                  alloc.rt_nn)
                        session.delete(alloc)

            # Add missing allocatable route targets to table
            for rt_nn in sorted(rt_nns):
                alloc = RouteTargetAllocation(rt_nn=rt_nn)
                session.add(alloc)



# vim: tabstop=4 shiftwidth=4 softtabstop=4
# encoding: utf-8

# Copyright 2014 Orange
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import random
import threading

from oslo_log import log as logging

from networking_bagpipe.bagpipe_bgp.common import looking_glass as lg
from networking_bagpipe.bagpipe_bgp.common import utils

LOG = logging.getLogger(__name__)

MAX_LABEL = 2**20-1


class LabelAllocator(lg.LookingGlassMixin):
    # Warning: does not reuse labels, will break if more than 2**20-16 labels
    # are requested along the life-cycle
    # FIXME: add code so that all the labels that were cleaned'up will *not*
    # be reused in a short time

    def __init__(self):
        # (Labels below 16 are reserved)
        self.current_label = random.randint(100, 200)
        # using a random start value will illustrate during demos and tests
        # that the label for a VRF does not
        # need be the same on all compute nodes
        self.labels = dict()

        self.lock = threading.Lock()

    @utils.synchronized
    def get_new_label(self, description):

        if self.current_label == MAX_LABEL+1:
            # Looking forward to the day will hit this one:
            LOG.error("All the 2^20 possible labels have been used at least "
                      "once, and this piece of code doesn't know how to reuse "
                      "them")
            raise Exception("Out of labels")

        label = self.current_label
        self.current_label += 1
        self.labels[label] = description

        LOG.debug("Allocated label %d for '%s'", label, description)
        return label

    @utils.synchronized
    def release(self, label):
        if label in self.labels:
            LOG.debug("released label %d ('%s')", label, self.labels[label])
            del self.labels[label]
        else:
            LOG.warn("asked to release a non registered label: %d", label)

    def get_lg_local_info(self, prefix):
        return self.labels



# vim: tabstop=4 shiftwidth=4 softtabstop=4
# encoding: utf-8

# Copyright 2014 Orange
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


class VPNNotFound(Exception):

    def __init__(self, vrf_id):
        self.message = "VRF %s could not be found" % vrf_id

    def __str__(self):
        return repr(self.message)


class MalformedIPAddress(Exception):

    def __init__(self, ip_address):
        self.message = "Address %s doesn't look valid" % ip_address

    def __str__(self):
        return repr(self.message)


class OVSBridgeNotFound(Exception):

    def __init__(self, bridge):
        self.message = "OVS bridge '%s' doesn't exist" % bridge

    def __str__(self):
        return str(self.message)


class OVSBridgePortNotFound(Exception):

    def __init__(self, interface, bridge):
        self.message = ("OVS Port %s doesn't exist on OVS Bridge %s" %
                        (interface, bridge))

    def __str__(self):
        return repr(self.message)


class RemotePEMACAddressNotFound(Exception):

    def __init__(self, ip_address):
        self.message = ("MAC address for %s could not be found. CAUTION:"
                        " Need direct MPLS/Eth connection" % ip_address)

    def __str__(self):
        return repr(self.message)


class APIException(Exception):
    pass



# Copyright (c) 2015 Orange.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import mock
from neutron_lib import context

from neutron.plugins.ml2 import config
from neutron.plugins.ml2 import managers
from neutron.plugins.ml2 import rpc
from neutron.tests.common import helpers
from neutron.tests.unit.plugins.ml2 import test_plugin
from neutron_lib.api.definitions import portbindings
from neutron_lib.api.definitions import provider_net as pnet

from networking_bagpipe.driver.type_route_target import TYPE_ROUTE_TARGET
import networking_bagpipe.rpc.client as bagpipe_rpc

from neutron_lib import constants as n_consts

HOST = 'bagpipe_host'
HOST_2 = HOST + '_2'
HOST_3 = HOST + '_3'

NOTIFIER = 'networking_bagpipe.rpc.client.BaGPipeAgentNotifyAPI'
DEVICE_OWNER_COMPUTE = 'compute:None'


def _get_linuxbridge_agent_dict(host, agent_type, binary,
                                bridge_mappings=None,
                                interface_mappings=None):
    agent = {
        'binary': binary,
        'host': host,
        'topic': n_consts.L2_AGENT_TOPIC,
        'agent_type': agent_type,
        'start_flag': True}

    if bridge_mappings is not None:
        agent['configurations']['bridge_mappings'] = bridge_mappings
    if interface_mappings is not None:
        agent['configurations']['interface_mappings'] = interface_mappings
    return agent


def register_linuxbridge_agent(host=HOST,
                               agent_type=n_consts.AGENT_TYPE_LINUXBRIDGE,
                               binary='bagpipe-linuxbridge-agent',
                               bridge_mappings=None,
                               interface_mappings=None):
    agent = _get_linuxbridge_agent_dict(host, agent_type, binary,
                                        bridge_mappings,
                                        interface_mappings)
    return helpers._register_agent(agent)


class TestBaGpipeRpcTestCase(test_plugin.Ml2PluginV2TestCase):
    _mechanism_drivers = ['bagpipe']

    def setUp(self):
        config.cfg.CONF.set_override('type_drivers',
                                     TYPE_ROUTE_TARGET,
                                     'ml2')
        config.cfg.CONF.set_override('tenant_network_types',
                                     TYPE_ROUTE_TARGET,
                                     'ml2',
                                     enforce_type=True)

        super(TestBaGpipeRpcTestCase, self).setUp()

        self.adminContext = context.get_admin_context()

        self.type_manager = managers.TypeManager()
        self.notifier = bagpipe_rpc.BaGPipeAgentNotifyAPI()
        self.callbacks = rpc.RpcCallbacks(self.notifier, self.type_manager)

        net_arg = {pnet.NETWORK_TYPE: TYPE_ROUTE_TARGET,
                   pnet.SEGMENTATION_ID: '101'}
        self._network = self._make_network(self.fmt, 'net1', True,
                                           arg_list=(pnet.NETWORK_TYPE,
                                                     pnet.SEGMENTATION_ID,),
                                           **net_arg)

        net_arg = {pnet.NETWORK_TYPE: TYPE_ROUTE_TARGET,
                   pnet.SEGMENTATION_ID: '102'}
        self._network2 = self._make_network(self.fmt, 'net2', True,
                                            arg_list=(pnet.NETWORK_TYPE,
                                                      pnet.SEGMENTATION_ID,),
                                            **net_arg)

        cast = ('networking_bagpipe.rpc.client.BaGPipeAgentNotifyAPI.'
                '_notification_host')
        cast_patch = mock.patch(cast)
        self.mock_cast = cast_patch.start()

        # don't wait for DHCP provisioning block to clear
        mock.patch('neutron.plugins.ml2.db.is_dhcp_active_on_any_subnet',
                   return_value=False).start()

    def _register_ml2_agents(self):
        register_linuxbridge_agent(host=HOST)
        register_linuxbridge_agent(host=HOST_2)
        register_linuxbridge_agent(host=HOST_3)

    def test_attach_port_on_bagpipe_network_called(self):
        self._register_ml2_agents()

        with self.subnet(network=self._network) as subnet:
            host_arg = {portbindings.HOST_ID: HOST}
            with self.port(subnet=subnet,
                           device_owner=DEVICE_OWNER_COMPUTE,
                           arg_list=(portbindings.HOST_ID,),
                           **host_arg) as port1:
                p1 = port1['port']
                s1 = subnet['subnet']
                device = 'tap' + p1['id']

                p1_ips = [p['ip_address'] for p in p1['fixed_ips']]
                s1_cidr = s1['cidr'][s1['cidr'].index('/'):]
                expected = {'id': p1['id'],
                            'network_id': p1['network_id'],
                            'mac_address': p1['mac_address'],
                            'ip_address': p1_ips[0] + s1_cidr,
                            'gateway_ip': s1['gateway_ip'],
                            'evpn': {'import_rt': '64512:101',
                                     'export_rt': '64512:101'
                                     }
                            }

                self.mock_cast.reset_mock()
                self.callbacks.update_device_up(self.adminContext,
                                                agent_id=HOST,
                                                device=device)

                self.mock_cast.assert_called_with(
                    mock.ANY, 'attach_port_on_bagpipe_network', expected,
                    HOST)

    def test_attach_port_on_bagpipe_network_called_two_agents(self):
        self._register_ml2_agents()

        with self.subnet(network=self._network) as subnet:
            host_arg = {portbindings.HOST_ID: HOST,
                        'admin_state_up': True}
            with self.port(subnet=subnet,
                           device_owner=DEVICE_OWNER_COMPUTE,
                           arg_list=(portbindings.HOST_ID, 'admin_state_up',),
                           **host_arg) as port1:
                host_arg = {portbindings.HOST_ID: HOST_2,
                            'admin_state_up': True}
                with self.port(subnet=subnet,
                               device_owner=DEVICE_OWNER_COMPUTE,
                               arg_list=(portbindings.HOST_ID,
                                         'admin_state_up',),
                               **host_arg) as port2:
                    p1 = port1['port']
                    s1 = subnet['subnet']
                    device1 = 'tap' + p1['id']

                    p1_ips = [p['ip_address'] for p in p1['fixed_ips']]
                    s1_cidr = s1['cidr'][s1['cidr'].index('/'):]
                    expected1 = {'id': p1['id'],
                                 'network_id': p1['network_id'],
                                 'mac_address': p1['mac_address'],
                                 'ip_address': p1_ips[0] + s1_cidr,
                                 'gateway_ip': s1['gateway_ip'],
                                 'evpn': {'import_rt': '64512:101',
                                          'export_rt': '64512:101'
                                          }
                                 }

                    p2 = port2['port']
                    device2 = 'tap' + p2['id']

                    p2_ips = [p['ip_address'] for p in p2['fixed_ips']]
                    expected2 = {'id': p2['id'],
                                 'network_id': p2['network_id'],
                                 'mac_address': p2['mac_address'],
                                 'ip_address': p2_ips[0] + s1_cidr,
                                 'gateway_ip': s1['gateway_ip'],
                                 'evpn': {'import_rt': '64512:101',
                                          'export_rt': '64512:101'
                                          }
                                 }

                    expected_calls = [
                        mock.call(mock.ANY, 'attach_port_on_bagpipe_network',
                                  expected1, HOST),
                        mock.call(mock.ANY, 'attach_port_on_bagpipe_network',
                                  expected2, HOST_2)
                    ]

                    self.mock_cast.reset_mock()
                    self.callbacks.update_device_up(self.adminContext,
                                                    agent_id=HOST,
                                                    device=device1)
                    self.callbacks.update_device_up(self.adminContext,
                                                    agent_id=HOST_2,
                                                    device=device2)

                    self.mock_cast.assert_has_calls(expected_calls)

    def test_attach_port_on_bagpipe_network_called_two_networks(self):
        self._register_ml2_agents()

        with self.subnet(network=self._network) as subnet:
            host_arg = {portbindings.HOST_ID: HOST_2}
            with self.port(subnet=subnet,
                           device_owner=DEVICE_OWNER_COMPUTE,
                           arg_list=(portbindings.HOST_ID,),
                           **host_arg) as port1:
                with self.subnet(network=self._network2,
                                 cidr='10.0.1.0/24') as subnet2:
                    host_arg = {portbindings.HOST_ID: HOST}
                    with self.port(subnet=subnet2,
                                   device_owner=DEVICE_OWNER_COMPUTE,
                                   arg_list=(portbindings.HOST_ID,),
                                   **host_arg) as port2:
                        p1 = port1['port']
                        s1 = subnet['subnet']
                        device1 = 'tap' + p1['id']

                        p1_ips = [p['ip_address'] for p in p1['fixed_ips']]
                        s1_cidr = s1['cidr'][s1['cidr'].index('/'):]
                        expected1 = {'id': p1['id'],
                                     'network_id': p1['network_id'],
                                     'mac_address': p1['mac_address'],
                                     'ip_address': p1_ips[0] + s1_cidr,
                                     'gateway_ip': s1['gateway_ip'],
                                     'evpn': {'import_rt': '64512:101',
                                              'export_rt': '64512:101'
                                              }
                                     }

                        p2 = port2['port']
                        s2 = subnet2['subnet']
                        device2 = 'tap' + p2['id']

                        p2_ips = [p['ip_address'] for p in p2['fixed_ips']]
                        s2_cidr = s2['cidr'][s2['cidr'].index('/'):]
                        expected2 = {'id': p2['id'],
                                     'network_id': p2['network_id'],
                                     'mac_address': p2['mac_address'],
                                     'ip_address': p2_ips[0] + s2_cidr,
                                     'gateway_ip': s2['gateway_ip'],
                                     'evpn': {'import_rt': '64512:102',
                                              'export_rt': '64512:102'
                                              }
                                     }

                        expected_calls = [
                            mock.call(mock.ANY,
                                      'attach_port_on_bagpipe_network',
                                      expected1, HOST_2),
                            mock.call(mock.ANY,
                                      'attach_port_on_bagpipe_network',
                                      expected2, HOST)
                        ]

                        self.mock_cast.reset_mock()
                        self.callbacks.update_device_up(self.adminContext,
                                                        agent_id=HOST_2,
                                                        device=device1)
                        self.callbacks.update_device_up(self.adminContext,
                                                        agent_id=HOST,
                                                        device=device2)

                        self.mock_cast.assert_has_calls(expected_calls)

    def test_delete_port(self):
        self._register_ml2_agents()

        with self.subnet(network=self._network) as subnet:
            host_arg = {portbindings.HOST_ID: HOST}
            with self.port(subnet=subnet,
                           device_owner=DEVICE_OWNER_COMPUTE,
                           arg_list=(portbindings.HOST_ID,),
                           **host_arg) as port:
                p1 = port['port']
                device1 = 'tap' + p1['id']

                self.mock_cast.reset_mock()
                self.callbacks.update_device_up(self.adminContext,
                                                agent_id=HOST,
                                                device=device1)

                with self.port(subnet=subnet,
                               device_owner=DEVICE_OWNER_COMPUTE,
                               arg_list=(portbindings.HOST_ID,),
                               **host_arg) as port2:
                    p2 = port2['port']
                    device2 = 'tap' + p2['id']

                    self.mock_cast.reset_mock()
                    self.callbacks.update_device_up(self.adminContext,
                                                    agent_id=HOST,
                                                    device=device2)

                self._delete('ports', port2['port']['id'])

                expected = {'id': p2['id'],
                            'network_id': p2['network_id'],
                            }

                self.mock_cast.assert_any_call(
                    mock.ANY, 'detach_port_from_bagpipe_network',
                    expected, HOST)

    def _update_and_check_portbinding(self, port_id, host_id):
        data = {'port': {'binding:host_id': host_id}}
        req = self.new_update_request('ports', data, port_id)
        res = self.deserialize(self.fmt,
                               req.get_response(self.api))
        self.assertEqual(host_id, res['port']['binding:host_id'])

    def _test_host_changed(self, twice):
        self._register_ml2_agents()
        helpers.register_dhcp_agent()

        with self.subnet(network=self._network) as subnet:
            host_arg = {portbindings.HOST_ID: HOST}
            with self.port(subnet=subnet,
                           device_owner=DEVICE_OWNER_COMPUTE,
                           arg_list=(portbindings.HOST_ID,),
                           **host_arg) as port1:
                p1 = port1['port']
                s1 = subnet['subnet']
                device1 = 'tap' + p1['id']

                p1_ips = [p['ip_address'] for p in p1['fixed_ips']]
                s1_cidr = s1['cidr'][s1['cidr'].index('/'):]
                expected1 = {'id': p1['id'],
                             'network_id': p1['network_id'],
                             'mac_address': p1['mac_address'],
                             'ip_address': p1_ips[0] + s1_cidr,
                             'gateway_ip': s1['gateway_ip'],
                             'evpn': {'import_rt': '64512:101',
                                      'export_rt': '64512:101'
                                      }
                             }

                self.callbacks.update_device_up(self.adminContext,
                                                agent_id=HOST,
                                                device=device1)

                self.mock_cast.assert_called_once_with(
                    mock.ANY, 'attach_port_on_bagpipe_network',
                    expected1, HOST)

                # FIXME(tmorin): migration code need to be updated
                # to follow the changes in Neutron commit
                # c5fa665de3173f3ad82cc3e7624b5968bc52c08d
#                 if twice:
#                     self._update_and_check_portbinding(p1['id'], HOST_3)
#                 self._update_and_check_portbinding(p1['id'], HOST_2)
#
#                 expected1 = {'id': p1['id'],
#                              'network_id': p1['network_id']}
#
#                 self.mock_cast.reset_mock()
#                 self.callbacks.get_device_details(self.adminContext,
#                                                   device=device1,
#                                                   agent_id=HOST_2)
#
#                 self.mock_cast.assert_called_once_with(
#                     mock.ANY, 'attach_port_on_bagpipe_network',
#                     expected1, HOST_2)
#
#                 self.mock_cast.assert_called_once_with(
#                     mock.ANY, 'detach_port_from_bagpipe_network',
#                     expected1, HOST)

    def test_host_changed(self):
        self._test_host_changed(twice=False)

    def test_host_changed_twice(self):
        self._test_host_changed(twice=True)



# Copyright (c) 2015 Orange.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_log import log as logging

import oslo_messaging

from neutron.common import rpc as n_rpc
from neutron.common import topics

LOG = logging.getLogger(__name__)

# until we have a better way to add something in the topic namespace
# from a python package external to Neutron...
topics_BAGPIPE = "bagpipe-l2"


class BaGPipeAgentNotifyAPI(object):
    """Base class for BaGPipe ML2 mech driver notification to agent RPC API"""

    def __init__(self):
        self.topic_bagpipe_update = topics.get_topic_name(topics.AGENT,
                                                          topics_BAGPIPE,
                                                          topics.UPDATE)

        target = oslo_messaging.Target(topic=topics.AGENT, version='1.0')
        self.client = n_rpc.get_client(target)

    def _notification_host(self, context, method, port_bagpipe_info, host):
        LOG.debug('Notify BaGPipe ML2 plugin agent %(host)s at %(topic)s '
                  'the message %(method)s with %(port_bagpipe_info)s',
                  {'host': host,
                   'topic': self.topic_bagpipe_update,
                   'method': method,
                   'port_bagpipe_info': port_bagpipe_info})

        cctxt = self.client.prepare(topic=self.topic_bagpipe_update,
                                    server=host)
        cctxt.cast(context, method, port_bagpipe_info=port_bagpipe_info)

    def attach_port_on_bagpipe_network(self, context, port_bagpipe_info,
                                       host=None):
        if host:
            self._notification_host(context,
                                    'attach_port_on_bagpipe_network',
                                    port_bagpipe_info,
                                    host)
        else:
            LOG.warning("attach_port_on_bagpipe_network called without"
                        " a host")

    def detach_port_from_bagpipe_network(self, context, port_bagpipe_info,
                                         host=None):
        if host:
            self._notification_host(context,
                                    'detach_port_from_bagpipe_network',
                                    port_bagpipe_info,
                                    host)
        else:
            LOG.warning("detach_port_from_bagpipe_network called without"
                        " a host")



# vim: tabstop=4 shiftwidth=4 softtabstop=4
# encoding: utf-8

# Copyright 2014 Orange
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import mock

from networking_bagpipe.bagpipe_bgp import constants as consts
from networking_bagpipe.bagpipe_bgp.vpn import manager
from networking_bagpipe.tests.unit.bagpipe_bgp import base as t


REDIRECTED_INSTANCE_ID1 = 'redirected-id1'
REDIRECTED_INSTANCE_ID2 = 'redirected-id2'


class TestVPNManager(t.TestCase):

    def setUp(self):
        super(TestVPNManager, self).setUp()
        mock.patch("networking_bagpipe.bagpipe_bgp.vpn.dataplane_drivers."
                   "instantiate_dataplane_drivers",
                   return_value={
                       'evpn': mock.Mock(),
                       'ipvpn': mock.Mock()
                   }).start()
        self.manager = manager.VPNManager.get_instance()

    def tearDown(self):
        super(TestVPNManager, self).tearDown()
        self.manager.stop()

    def test_redirect_traffic_single_instance(self):
        redirect_instance = self.manager.redirect_traffic_to_vpn(
            REDIRECTED_INSTANCE_ID1, consts.IPVPN, t._rt_to_string(t.RT5)
        )

        # Check some VPN manager and redirect instance lists consistency
        self.assertIn(
            manager.redirect_instance_extid(consts.IPVPN,
                                            t._rt_to_string(t.RT5)),
            self.manager.vpn_instances)
        self.assertIn(REDIRECTED_INSTANCE_ID1,
                      redirect_instance.redirected_instances)

    def test_redirect_traffic_multiple_instance(self):
        redirect_instance_1 = self.manager.redirect_traffic_to_vpn(
            REDIRECTED_INSTANCE_ID1, consts.IPVPN, t._rt_to_string(t.RT5)
        )
        redirect_instance_2 = self.manager.redirect_traffic_to_vpn(
            REDIRECTED_INSTANCE_ID2, consts.IPVPN, t._rt_to_string(t.RT5)
        )

        # Check that same redirect instance is returned
        self.assertEqual(redirect_instance_2, redirect_instance_1)
        # Check some VPN manager and redirect instance lists consistency
        self.assertIn(
            manager.redirect_instance_extid(consts.IPVPN,
                                            t._rt_to_string(t.RT5)),
            self.manager.vpn_instances)
        self.assertIn(REDIRECTED_INSTANCE_ID1,
                      redirect_instance_1.redirected_instances)
        self.assertIn(REDIRECTED_INSTANCE_ID2,
                      redirect_instance_1.redirected_instances)

    def test_stop_redirect_traffic_multiple_instance(self):
        redirect_instance = self.manager.redirect_traffic_to_vpn(
            REDIRECTED_INSTANCE_ID1, consts.IPVPN, t._rt_to_string(t.RT5)
        )
        self.manager.redirect_traffic_to_vpn(
            REDIRECTED_INSTANCE_ID2, consts.IPVPN, t._rt_to_string(t.RT5)
        )

        # Check some VPN manager and redirect instance lists consistency
        self.manager.stop_redirect_to_vpn(REDIRECTED_INSTANCE_ID2,
                                          consts.IPVPN, t._rt_to_string(t.RT5))
        self.assertNotIn(REDIRECTED_INSTANCE_ID2,
                         redirect_instance.redirected_instances)

        self.manager.stop_redirect_to_vpn(REDIRECTED_INSTANCE_ID1,
                                          consts.IPVPN, t._rt_to_string(t.RT5))
        self.assertTrue(not self.manager.vpn_instances)
